import dagster
import igraph as ig
import numpy as np
import pandas as pd

from ireiat.config.constants import (
    INTERMEDIATE_DIRECTORY_ARGS,
)
from ireiat.config.data_pipeline import TAPNetworkConfig, TAPRailConfig
from ireiat.config.rail_enum import EdgeType
from ireiat.data_pipeline.metadata import publish_metadata


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json({"index": False}),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def tap_highway_network_dataframe(
    context: dagster.AssetExecutionContext,
    highway_network_dataframe: pd.DataFrame,
    config: TAPNetworkConfig,
) -> pd.DataFrame:
    """Entire highway network to represent the TAP, complete with capacity and cost information"""
    # fill out other fields needed for the TAP
    tap_network = highway_network_dataframe
    tap_network["speed"] = tap_network["speed"].fillna(
        tap_network["speed"].mean()
    )  # fill in any null speeds
    # replace any zero speeds with the mean speed
    tap_network["speed"] = tap_network["speed"].replace(0, tap_network["speed"].mean())
    tap_network["fft"] = tap_network["length"] / tap_network["speed"]
    tap_network["beta"] = config.default_network_beta
    tap_network["alpha"] = config.default_network_alpha
    tap_network["capacity"] = config.default_capacity_ktons
    tap_network = tap_network.sort_values(["tail", "head"])

    assert tap_network["speed"].min() > 0
    context.log.info(f"TAP highway network dataframe created with {len(tap_network)} edges.")
    publish_metadata(context, tap_network)
    return tap_network


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json({"index": False}),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def tap_rail_network_dataframe(
    context: dagster.AssetExecutionContext,
    rail_graph_with_county_connections: ig.Graph,
    config: TAPRailConfig,
) -> pd.DataFrame:
    """Entire rail network to represent the TAP, complete with capacity and cost information"""
    # fill out other fields needed for the TAP

    connected_edge_tuples = [
        (e.source, e.target, e["length"], e["edge_type"], e["owners"], e["speed"], e["tracks"])
        for e in rail_graph_with_county_connections.es
    ]

    # create and return a dataframe
    tap_network = pd.DataFrame(
        connected_edge_tuples,
        columns=["tail", "head", "length", "edge_type", "owners", "speed", "tracks"],
    )

    tap_network["speed"] = tap_network["speed"].fillna(config.default_speed_mph)
    tap_network["length"] = tap_network["length"].fillna(0.1)
    tap_network["fft"] = tap_network["length"] / tap_network["speed"]

    # Apply beta and alpha based on edge type
    is_im_capacity_edge = tap_network["edge_type"] == EdgeType.IM_CAPACITY.value
    is_rail_link_edge = tap_network["edge_type"] == EdgeType.RAIL_LINK.value

    tap_network["capacity"] = np.where(
        is_im_capacity_edge,
        config.intermodal_facility_capacity_ktons,
        np.where(
            is_rail_link_edge,
            tap_network["tracks"] * config.capacity_ktons_per_track,
            config.capacity_ktons_default,
        ),
    )
    tap_network["beta"] = np.where(
        is_im_capacity_edge,
        config.intermodal_edge_beta,
        config.default_network_beta,
    )
    tap_network["alpha"] = np.where(
        is_im_capacity_edge,
        config.intermodal_edge_alpha,
        config.default_network_alpha,
    )

    tap_network = tap_network.sort_values(["tail", "head"])

    assert tap_network["speed"].min() > 0
    context.log.info(f"TAP rail network dataframe created with {len(tap_network)} edges.")
    publish_metadata(context, tap_network)
    return tap_network


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json({"index": False}),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def tap_marine_network_dataframe(
    context: dagster.AssetExecutionContext,
    marine_network_dataframe: pd.DataFrame,
    config: TAPNetworkConfig,
) -> pd.DataFrame:
    """Entire marine network to represent the TAP, complete with capacity and cost information"""
    # fill out other fields needed for the TAP
    tap_network = marine_network_dataframe

    if "speed" not in tap_network.columns:
        # If the 'speed' column doesn't exist
        tap_network["speed"] = config.default_speed_mph
    else:
        # Fill missing values with the default value
        tap_network["speed"] = tap_network["speed"].fillna(config.default_speed_mph)

    tap_network["fft"] = tap_network["length"] / tap_network["speed"]
    tap_network["beta"] = config.default_network_beta
    tap_network["alpha"] = config.default_network_alpha
    tap_network["capacity"] = config.default_capacity_ktons
    tap_network = tap_network.sort_values(["tail", "head"])

    assert tap_network["speed"].min() > 0
    context.log.info(f"TAP marine network dataframe created with {len(tap_network)} edges.")
    publish_metadata(context, tap_network)
    return tap_network



import dagster

from ireiat.data_pipeline.io_manager import asset_spec_factory

narn_links_spec = dagster.AssetSpec(
    key=dagster.AssetKey("narn_rail_network_links_spec"),
    description="Publicly available GIS data for railway network links",
    metadata={
        "format": "zip",
        "filename": "narn_rail_links.zip",
        "source_path": "raw/",
        "dashboard_url": dagster.MetadataValue.url(
            "https://hub.arcgis.com/api/download/v1/items/e143f436d4774402aa8cca1e663b1d24/shapefile?redirect=true&layers=0"
        ),
    },
)
narn_links_src = asset_spec_factory(narn_links_spec)

intermodal_terminals_spec = dagster.AssetSpec(
    key=dagster.AssetKey("intermodal_terminals_spec"),
    description="XLSX containing intermodal terminal information",
    metadata={
        "format": "xlsx",
        "filename": "im_terminals.xlsx",
        "source_path": "raw/",
        "dashboard_url": dagster.MetadataValue.url(
            "https://hub.arcgis.com/api/download/v1/items/e6332597fcd546d3899f60ec9b9157b9/excel?redirect=true&layers=0"
        ),
    },
)
intermodal_terminals_src = asset_spec_factory(intermodal_terminals_spec)



from collections import defaultdict
from copy import deepcopy
from typing import Dict, Tuple, List

import dagster
import geopandas
import numpy as np


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def faf_id_to_county_areas(
    context: dagster.AssetExecutionContext,
    faf5_regions_src: geopandas.GeoDataFrame,
    us_county_shp_files_src: geopandas.GeoDataFrame,
) -> Dict[str, Dict[Tuple[str, str], float]]:
    """Reads FAF5 GIS data, county GIS data, and computes a FAF zone -> (County FIPS, State FIPS) -> % area map"""
    # area computation requires consistent CRS representation
    county_gdf = us_county_shp_files_src.to_crs(faf5_regions_src.crs)
    county_gdf["area"] = county_gdf.area

    # county shp file data is unique at STATEFP, COUNTYFP composite key
    assert len(county_gdf[["STATEFP", "COUNTYFP"]].drop_duplicates()) == len(county_gdf)

    # construct a map of this dataframe's index to (STATEFP,COUNTYFP)
    county_idx_map = {row.Index: (row.STATEFP, row.COUNTYFP) for row in county_gdf.itertuples()}

    # Compute which counties are part of which FAF5 region (and what percentage of their area)
    INTERSECTION_AREA_TOLERANCE_PERCENT = 0.01
    faf_zone_to_county = dict()
    for row in faf5_regions_src.itertuples():
        # return just the "overlap" of the FAF5 region with county data
        county_intersection = county_gdf.intersection(row.geometry)
        nonzero_county_intersection = county_intersection[~county_intersection.is_empty]

        # get the full county data and compute the percentage of county overlap by area,
        # ignoring anything less than the tolerance
        full_intersected_counties = county_gdf.loc[nonzero_county_intersection.index]
        intersected_counties_relative_area = (
            nonzero_county_intersection.area / full_intersected_counties.area
        )
        tol_adjusted_county_intersection = intersected_counties_relative_area.loc[
            np.where(
                intersected_counties_relative_area > INTERSECTION_AREA_TOLERANCE_PERCENT,
                True,
                False,
            )
        ]

        faf_zone_to_county[row.FAF_Zone] = np.round(tol_adjusted_county_intersection, 2).to_dict()

    # compute a map that links faf_zone_ids to counties
    # compute a map that tabulates the percentage of a county's area that's covered by a FAF zone
    county_coverage: Dict[Tuple[str, str], float] = defaultdict(float)
    county_id_to_faf_zone_id: Dict[Tuple[str, str], List[str]] = defaultdict(list)
    for faf_zone_id, val in faf_zone_to_county.items():
        for county_id, coverage in val.items():
            county_coverage[county_id] += coverage
            county_id_to_faf_zone_id[county_id].append(faf_zone_id)

    # get the county keys that are not totally covered
    not_totally_covered_county_ids = [
        k for k, covered_area_pct in county_coverage.items() if covered_area_pct < 1
    ]

    # if a county's area has not been totally covered AND there's more than one FAF zone that covers this county...we have a problem
    assert {
        k: len(v)
        for k, v in county_id_to_faf_zone_id.items()
        if k in not_totally_covered_county_ids and len(v) > 1
    } == dict()

    # as long as there is only 1 FAF region covering a "problem" county, we assign *all* of the county's area to be fully part of the FAF region
    faf_zone_to_county_canonical = deepcopy(faf_zone_to_county)
    for faf_zone, county_dict in faf_zone_to_county_canonical.items():
        for some_uncovered_county_id in not_totally_covered_county_ids:
            if some_uncovered_county_id in county_dict.keys():
                county_dict[some_uncovered_county_id] = 1.0

    county_coverage_canonical: Dict[Tuple[str, str], float] = defaultdict(float)
    for faf_zone_id, val in faf_zone_to_county_canonical.items():
        for county_id, coverage in val.items():
            county_coverage_canonical[county_id] += coverage

    # ensure that our canonical linkage of FAF zone IDs to county areas is complete
    assert min(county_coverage_canonical.values()) == 1.0

    # ensure every faf zone is covered
    assert (set(faf5_regions_src["FAF_Zone"]) - set(faf_zone_to_county_canonical.keys())) == set()

    # we relied on the index of the county geographic data to identify each county row - but, for usage with census data,
    # we serialize using the (STATEFP, COUNTYFP) code
    faf_zone_id_to_state_county_id = dict()
    for faf_zone_id, county_data in faf_zone_to_county_canonical.items():
        faf_zone_id_to_state_county_id[faf_zone_id] = {
            county_idx_map.get(k): v for k, v in county_data.items()
        }

    # bubble up metadata associated with this result
    context.add_output_metadata({"records": len(faf_zone_id_to_state_county_id.keys())})
    return faf_zone_id_to_state_county_id



from enum import IntEnum


class FAFMode(IntEnum):
    """
    FAF transit mode. See https://www.bts.gov/sites/bts.dot.gov/files/2021-02/FAF5-User-Guide.pdf
    """

    TRUCK = 1
    RAIL = 2
    WATER = 3
    AIR = 4
    MULTIPLE_AND_MAIL = 5
    PIPELINE = 6
    OTHER_AND_UNKNOWN = 7
    NO_DOMESTIC_MODE = 8



import logging
import os
import pickle
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import pyogrio

from ireiat.config.constants import CACHE_PATH

logger = logging.getLogger(__name__)


class PostProcessor:
    """Generate output artifacts from TAP solution files"""

    _tap_solution_path: Path
    _strongly_connected_graph_path: Path
    _shp_file_path: Path
    _solution_output_artifacts: Path = None
    _congestion_png_path: Path = None

    @property
    def artifact_output_path(self) -> Path:
        """Location in local filesystem for solution output artifacts"""
        if self._solution_output_artifacts is None:
            self._solution_output_artifacts = CACHE_PATH / "solution_output_artifacts"
            os.makedirs(self._solution_output_artifacts, exist_ok=True)
        return self._solution_output_artifacts

    @property
    def congestion_png_filename(self) -> str:
        return self._tap_solution_path.stem + ".png"

    def __init__(
        self,
        tap_solution_path: Path,
        strongly_connected_graph_path: Path,
        shp_file_path: Path,
    ):
        self._tap_solution_path = Path(tap_solution_path)
        self._strongly_connected_graph_path = strongly_connected_graph_path
        self._shp_file_path = shp_file_path

    def _generate_congestion_png(self):
        logger.info("Reading solution and graph data")
        traffic = pd.read_parquet(self._tap_solution_path)
        for cast_to_int_column in ["from", "to"]:
            traffic[cast_to_int_column] = traffic[cast_to_int_column].astype(int)

        # load the graph which tracks original faf_link_ids
        with open(self._strongly_connected_graph_path, "rb") as fp:
            strongly_connected_graph = pickle.load(fp)

        # we need to be careful about the order here. the solution (of edges) may be in a different order,
        # so we need to get the original_id of the edge on the network...and then join that data
        # into the solution
        edge_to_shp_file_link_id = {
            (es.source, es.target): es["original_id"] for es in strongly_connected_graph.es
        }

        # map the edge (source_vertex, destination_vertex) to the original shp edge id, which is stored in the graph
        shp_link_ids = [
            edge_to_shp_file_link_id[(row._1, row.to)]
            for row in traffic[["from", "to"]].itertuples()
        ]

        traffic["shp_link_id"] = shp_link_ids

        # this is a bit of a fudge in that the utilization on the same road (2 way) could be far above the capacity...
        # ideally we would set the capacity on each directed segment and sum the flows and capacities and then divide...
        logger.info("Computing utilization")
        traffic["utilization"] = traffic["flow"] / traffic["capacity"]
        grouped_traffic = traffic.groupby("shp_link_id")[["utilization"]].sum()

        # read the network
        logger.info("Reading network data")
        shp_file_gdf = pyogrio.read_dataframe(self._shp_file_path, use_arrow=True)
        if "id" not in shp_file_gdf.columns:
            if "FRAARCID" in shp_file_gdf.columns:
                shp_file_gdf["id"] = shp_file_gdf["FRAARCID"]
                logger.info("Assuming running rail postprocessing")
            else:
                shp_file_gdf["id"] = shp_file_gdf.index
                logger.info("Assuming running marine postprocessing")
        else:
            shp_file_gdf = shp_file_gdf.rename({"ID": "id"}, axis="columns")
            logger.info("Assuming running highway postprocessing")
        flows_with_geometry = (
            shp_file_gdf[["id", "geometry"]].set_index("id").join(grouped_traffic, how="left")
        )
        flows_with_geometry["utilization"] = flows_with_geometry["utilization"].fillna(0)

        # all grouped traffic (and associated edges) are found within the FAF data
        # assert len(flows_with_geometry) == len(grouped_traffic)

        non_zero_utilization = flows_with_geometry.loc[flows_with_geometry["utilization"] > 0]
        logger.info(non_zero_utilization.describe())
        logger.info("Generating plots for congestion")
        fig, ax = plt.subplots(figsize=(15, 8))
        flows_with_geometry.plot(ax=ax, color="grey", alpha=0.2)
        non_zero_utilization.plot(ax=ax, colors=plt.cm.YlOrRd(non_zero_utilization["utilization"]))
        ax.set_xlim(-130, -60)
        ax.set_ylim(23, 50)
        plt.tight_layout()
        plt.savefig(self.artifact_output_path / self.congestion_png_filename)
        logger.info(f"Plots saved to {self.artifact_output_path}")

    def postprocess(self):
        # create a graph of the traffic image
        self._generate_congestion_png()



from typing import Dict, Tuple

import dagster
import geopandas
import igraph as ig
import pandas as pd

from ireiat.config.constants import INTERMEDIATE_DIRECTORY_ARGS, ALBERS_CRS, METERS_PER_MILE
from ireiat.data_pipeline.metadata import publish_metadata
from ireiat.util.graph import (
    get_coordinates_from_geoframe,
    generate_zero_based_node_maps,
    get_allowed_node_indices,
)


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={"format": "parquet", **INTERMEDIATE_DIRECTORY_ARGS},
)
def undirected_marine_edges(
    context: dagster.AssetExecutionContext, marine_network_links_src: geopandas.GeoDataFrame
) -> pd.DataFrame:
    """For each undirected edge in the marine dataset, create a row in the table with origin_lat, origin_long,
    destination_lat, and destination_long, along with several other edge fields of interest"""

    link_coords = get_coordinates_from_geoframe(marine_network_links_src)
    marine_network_albers = marine_network_links_src.to_crs(ALBERS_CRS)
    distance_miles = marine_network_albers.geometry.length / METERS_PER_MILE
    link_coords = pd.concat([link_coords, pd.Series(distance_miles, name="distance_miles")], axis=1)
    publish_metadata(context, link_coords)
    return link_coords


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def strongly_connected_marine_graph(
    context: dagster.AssetExecutionContext, undirected_marine_edges: pd.DataFrame
) -> ig.Graph:
    """iGraph object representing a strongly connected, directed graph based on the marine network"""
    edge_tuples = []
    edge_attributes = []
    complete_marine_node_to_idx: Dict[Tuple[float, float], int] = generate_zero_based_node_maps(
        undirected_marine_edges
    )

    for idx, row in enumerate(undirected_marine_edges.itertuples()):
        origin_coords = (row.origin_latitude, row.origin_longitude)
        destination_coords = (row.destination_latitude, row.destination_longitude)
        tail, head = (
            complete_marine_node_to_idx[origin_coords],
            complete_marine_node_to_idx[destination_coords],
        )
        # Add edges in both directions
        edge_tuples.append((tail, head))
        edge_tuples.append((head, tail))

        # Duplicate attributes for both directions
        edge_attributes.append((row.distance_miles, idx, origin_coords, destination_coords))
        edge_attributes.append((row.distance_miles, idx, destination_coords, origin_coords))

    # generate a graph from all nodes
    n_vertices = len(complete_marine_node_to_idx)
    context.log.info(f"Original number of nodes {n_vertices}, edges {len(edge_tuples)}.")
    g = ig.Graph(
        n_vertices,
        edge_tuples,
        edge_attrs={
            "length": [attr[0] for attr in edge_attributes],
            "original_id": [attr[1] for attr in edge_attributes],
            "origin_coords": [attr[2] for attr in edge_attributes],
            "destination_coords": [attr[3] for attr in edge_attributes],
        },
        directed=True,
    )
    context.log.info(f"Initial constructed graph connected?: {g.is_connected()}")
    allowed_node_indices = get_allowed_node_indices(g)

    # construct a connected subgraph
    connected_subgraph = g.subgraph(allowed_node_indices)
    return connected_subgraph


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={"format": "parquet", **INTERMEDIATE_DIRECTORY_ARGS},
)
def marine_network_dataframe(
    context: dagster.AssetExecutionContext, strongly_connected_marine_graph: ig.Graph
) -> pd.DataFrame:
    """Returns a dataframe of graph edges along with attributes needed to solve the TAP"""
    connected_edge_tuples = [
        (e.source, e.target, e["length"], *e["origin_coords"], *e["destination_coords"])
        for e in strongly_connected_marine_graph.es
    ]

    # create and return a dataframe
    pdf = pd.DataFrame(
        connected_edge_tuples,
        columns=[
            "tail",
            "head",
            "length",
            "origin_latitude",
            "origin_longitude",
            "destination_latitude",
            "destination_longitude",
        ],
    )
    context.log.info(f"Marine network dataframe created with {len(pdf)} edges.")
    publish_metadata(context, pdf)
    return pdf



import datetime
import logging
from pathlib import Path

import requests
from tqdm import tqdm

logger = logging.getLogger(__name__)


def download_uncached_file(url: str, save_path: Path | str, force: bool = False) -> bool:
    """Simple download helper to stream the contents of `url` to `save_path`.
    If `save_path` does not exist (including parent directories) on the filesystem, it will be created.

    After downloading, this method creates a "{save_path}.sentinel" file to indicate that the
    download was complete. If the sentinel file is not present, the download is assumed
    to have been partial and the file is re-downloaded.
    """

    # assume save_path is a fully qualified filename x/y/z.ending
    if isinstance(save_path, str):
        save_path = Path(save_path)
    save_path.parent.mkdir(parents=True, exist_ok=True)
    sentinel_file_path = save_path.parent / f"{save_path.name}.sentinel"

    if not force:
        if save_path.exists():
            if sentinel_file_path.exists():
                logger.info(f"File {save_path} and sentinel exists - skipping download.")
                return True
            else:
                logger.info("Sentinel not found! Re-downloading file...")

    # remove the file and its sentinel
    save_path.unlink(missing_ok=True)
    sentinel_file_path.unlink(missing_ok=True)
    response = requests.get(
        url, stream=True, verify=False, timeout=10
    )  # 10 second timeout for connect or read
    total_size_in_bytes = int(response.headers.get("content-length", 0))
    block_size = 1024  # 1 kb
    progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)
    with open(save_path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    with open(sentinel_file_path, "w") as fp:
        fp.write(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

    return True



from itertools import chain, product
from typing import Tuple, Set, Dict, Any

import igraph as ig

from ireiat.config.rail_enum import EdgeType


def _generate_subgraphs(g: ig.Graph, separation_attribute: str = "owner") -> Dict[str, ig.Graph]:
    """
    Returns subgraphs by unique entries in the `separation_attribute`  for the graph.
    Returned subgraphs have an 'original_idx' field with the format {X}{Y}, where X is the
    `separation_attribute` string and Y is the original vertex index.
    E.g. 0->1 (Graph A) would have ['A0','A1'] vertex attributes

    Additionally, each subgraph's `separation_attribute` now takes the string of the attribute value
    to which the subgraph belongs. E.g. Graph with 0->1 (separation_attribute={'A','B'} would have
    two graphs returned, each with `separation_attribute='A'` or `separation_attribute='B'`.

    :param g: graph to separate by a given attribute
    :return: Dict of subgraphs, keyed by unique separation attribute values"""
    # create subgraphs by owner
    unique_owners_across_edges = set(chain.from_iterable(g.es[separation_attribute]))
    edge_list_by_owner: Dict[str, Set[int]] = {k: set() for k in unique_owners_across_edges}
    for e in g.es:
        for k in e[separation_attribute]:
            edge_list_by_owner[k].add(e.index)
    # retain all vertices in the subgraph to preserve numbering
    subgraphs = {k: g.subgraph_edges(v) for k, v in edge_list_by_owner.items()}
    for k, subgraph in subgraphs.items():
        subgraph.es[separation_attribute] = [f"{k}" for _ in subgraph.es]
    return subgraphs


def _generate_impedances(
    g: ig.Graph, separation_attribute="owner"
) -> Set[Tuple[str, str, str, str]]:
    """
    Given a graph with a `separation_attribute` on each edge, determine the impedance
    edges that would be needed to join subgraphs that were created from unique values of the
    separation attribute. For example, if a graph of 0 -> 1 -> 2 had 'owners' 'A' on the first
    edge and 'B' on the second edge (along with origin and destination coords for each edge),
    this method would return `{('A', destination_coords, 'B', origin_coords)}` given
    that a graph split by owner would be A graph: 0->1 and B graph: 1->2 and there would be an
    impedance edge between 'A1' and 'B1' (with appropriate coordinates returned).

    :param g: graph to generate impedance edges from
    :param separation_attribute: string identifer on the graph edges
    :return: Set of tuples of impedance graph edges
    """
    impedances = set()
    for v in g.vs:
        in_edges, out_edges = v.in_edges(), v.out_edges()
        if len(out_edges) == 1 and len(in_edges) == 1:
            in_owners, out_owners = (
                in_edges[0][separation_attribute],
                out_edges[0][separation_attribute],
            )
            if in_owners == out_owners:
                continue  # we have single in/out edges with the same owner on each side
            else:
                # we have single in/out edges with different owners
                in_edge_destination_coords = in_edges[0]["destination_coords"]
                out_edge_origin_coords = out_edges[0]["origin_coords"]
                for in_edge_owner, out_edge_owner in product(in_owners, out_owners):
                    if in_edge_owner != out_edge_owner:
                        impedances.add(
                            (
                                in_edge_owner,
                                in_edge_destination_coords,
                                out_edge_owner,
                                out_edge_origin_coords,
                            )
                        )
        else:
            for in_edge in in_edges:
                in_edge_destination_coords = in_edge["destination_coords"]
                in_edge_owners = in_edge[separation_attribute]
                for out_edge in out_edges:
                    out_edge_origin_coords = out_edge["origin_coords"]
                    out_edge_owners = out_edge[separation_attribute]
                    for in_edge_owner, out_edge_owner in product(in_edge_owners, out_edge_owners):
                        if in_edge_owner != out_edge_owner:
                            impedances.add(
                                (
                                    in_edge_owner,
                                    in_edge_destination_coords,
                                    out_edge_owner,
                                    out_edge_origin_coords,
                                )
                            )

    return impedances


def generate_impedance_graph(g: ig.Graph, separation_attribute="owner") -> ig.Graph:
    """Given a graph whose edges all contain `separation_attribute`, return an "exploded"
    graph with impedances edges between vertices that have different values of the
    separation attribute. See the detailed test cases for how this method is intended to function.

    :param g: iGraph with `separation_attribute` of type `Set`
    :param separation_attribute: string identifying the attribute
    :return: an exploded graph with impedance edges

    """
    impedances = _generate_impedances(g, separation_attribute=separation_attribute)
    print(f"Generated {len(impedances)} impedances")
    subgraphs = _generate_subgraphs(g, separation_attribute=separation_attribute)
    print(f"Generated {len(subgraphs)} subgraphs")
    disjoint_union = ig.disjoint_union(subgraphs.values())

    # cache vertices from/to in the disjoint graph to facilitate fast lookups
    vertices_from = {
        (e["owners"], e["destination_coords"]): e.target_vertex.index for e in disjoint_union.es
    }
    vertices_to = {
        (e["owners"], e["origin_coords"]): e.source_vertex.index for e in disjoint_union.es
    }
    impedance_edges = [
        (vertices_from[(src_owner, dest_coords)], vertices_to[(dest_owner, origin_coords)])
        for src_owner, dest_coords, dest_owner, origin_coords in impedances
    ]

    # construct default edge attributes for impedance edges and add them
    impedance_edge_attrs: Dict[str, Any] = dict()
    impedance_edge_attrs["edge_type"] = [EdgeType.IMPEDANCE_LINK.value for _ in impedance_edges]
    impedance_edge_attrs[separation_attribute] = ["imp" for _ in impedance_edges]
    impedance_edge_attrs["length"] = [250 for _ in impedance_edges]
    disjoint_union.add_edges(impedance_edges, impedance_edge_attrs)

    # eliminate zero degree vertices, preserved when creating subgraphs
    disjoint_union.delete_vertices(disjoint_union.vs.select(_degree=0))
    return disjoint_union



from collections import defaultdict
from itertools import chain
from typing import Dict, Tuple, Any

import dagster
import geopandas
import igraph as ig
import numpy as np
import pandas as pd
from sklearn.neighbors import BallTree

from ireiat.config.constants import INTERMEDIATE_DIRECTORY_ARGS, RR_MAPPING
from ireiat.data_pipeline.assets.rail_network.impedance import generate_impedance_graph
from ireiat.data_pipeline.metadata import publish_metadata
from ireiat.util.graph import (
    get_coordinates_from_geoframe,
    generate_zero_based_node_maps,
    get_allowed_node_indices,
)
from ireiat.config.rail_enum import EdgeType, VertexType

SEPARATION_ATTRIBUTE_NAME: str = "owners"  # field used to represent owners and trackage rights


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={"format": "parquet", "use_geopandas": True, **INTERMEDIATE_DIRECTORY_ARGS},
)
def filtered_and_processed_rail_network_links(
    context: dagster.AssetExecutionContext, narn_rail_network_links_src: geopandas.GeoDataFrame
) -> geopandas.GeoDataFrame:
    """Preprocess the rail links data by filtering records that are known to be defunct (e.g. NET=A, Abandoned)
    and consolidating the railroad owner(s) and railroad trackage rights into a single field"""
    # remove abandoned, physically removed, and out of service
    EXCLUDED_TRACK_NET_VALUES = ["A", "R", "X"]
    bad_track_filter = narn_rail_network_links_src["NET"].isin(EXCLUDED_TRACK_NET_VALUES)

    # Remove links that have AMTK as the owner with no other owner and all TRKRGHTS columns are null
    amtk_filter = (
        (narn_rail_network_links_src["RROWNER1"] == "AMTK")
        & narn_rail_network_links_src["RROWNER2"].isna()
        & narn_rail_network_links_src["RROWNER3"].isna()
    )
    for col in [c for c in narn_rail_network_links_src.columns if "TRKRGHTS" in c]:
        amtk_filter &= narn_rail_network_links_src[col].isna()

    real_lines = narn_rail_network_links_src[~amtk_filter & ~bad_track_filter].copy()
    ownership_cols = [col for col in real_lines.columns if "RROWNER" in col or "TRKRGHTS" in col]

    for col in ownership_cols:
        real_lines[col] = real_lines[col].replace(RR_MAPPING)

    owner_set = real_lines[ownership_cols].apply(lambda x: set(filter(pd.notna, x)), axis=1)
    # remove AMTK as a relevant owner, since interchange costs will not matter with AMTK
    owner_set.apply(lambda x: x.discard("AMTK"))
    # Add CSXT and NS to OWNERS if PAS is one of the owners (PAS is jointly owned by CSXT and NS)
    owner_set.apply(lambda x: x.update(["CSXT", "NS"]) if "PAS" in x else x)

    real_lines[SEPARATION_ATTRIBUTE_NAME] = owner_set
    real_lines["TRACKS"] = real_lines["TRACKS"].replace(0, 1)

    # columns to retain
    cols_to_retain = ["FRAARCID", "MILES", "TRACKS", SEPARATION_ATTRIBUTE_NAME, "geometry"]
    # exclude items with no owners at all
    real_lines_with_owners = real_lines.loc[
        real_lines[SEPARATION_ATTRIBUTE_NAME].apply(len) >= 1, cols_to_retain
    ].copy()
    context.log.info(
        f"Rail links data loaded and preprocessed with {len(real_lines_with_owners)} rail links"
    )
    publish_metadata(context, real_lines_with_owners)
    return real_lines_with_owners


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={"format": "parquet", **INTERMEDIATE_DIRECTORY_ARGS},
)
def undirected_rail_edges(
    context: dagster.AssetExecutionContext,
    filtered_and_processed_rail_network_links: geopandas.GeoDataFrame,
) -> pd.DataFrame:
    """For each undirected edge in the rail dataset, create a row in the table with origin_lat, origin_long,
    destination_lat, and destination_long, along with several other edge fields of interest"""

    link_coords = get_coordinates_from_geoframe(filtered_and_processed_rail_network_links)
    fields_to_retain = ["FRAARCID", SEPARATION_ATTRIBUTE_NAME, "MILES", "TRACKS"]
    link_coords = pd.concat(
        [filtered_and_processed_rail_network_links[fields_to_retain], link_coords], axis=1
    )  # join in the direction
    publish_metadata(context, link_coords)
    return link_coords


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def strongly_connected_rail_graph(
    context: dagster.AssetExecutionContext, undirected_rail_edges: pd.DataFrame
) -> ig.Graph:
    """iGraph object representing a strongly connected, directed graph based on the rail network"""
    # generate directed edges from the undirected edges based on the "dir" field
    edge_tuples = []
    edge_attributes = []
    undirected_rail_edges[SEPARATION_ATTRIBUTE_NAME] = undirected_rail_edges[
        SEPARATION_ATTRIBUTE_NAME
    ].apply(set)

    complete_rail_node_to_idx: Dict[Tuple[float, float], int] = generate_zero_based_node_maps(
        undirected_rail_edges
    )
    for row in undirected_rail_edges.itertuples():
        origin_coords = (row.origin_latitude, row.origin_longitude)
        destination_coords = (row.destination_latitude, row.destination_longitude)
        tail, head = (
            complete_rail_node_to_idx[origin_coords],
            complete_rail_node_to_idx[destination_coords],
        )

        # record some original edge information needed for visualization and/or TAP setup
        attribute_tuple = (
            row.MILES,
            row.FRAARCID,
            row.owners,
            origin_coords,
            destination_coords,
            row.TRACKS,
        )
        edge_tuples.append((tail, head))
        edge_attributes.append(attribute_tuple)
        edge_tuples.append((head, tail))
        edge_attributes.append(attribute_tuple)

    # generate a graph from all nodes
    n_vertices = len(complete_rail_node_to_idx)
    context.log.info(f"Original number of nodes {n_vertices}, edges {len(edge_tuples)}.")
    g = ig.Graph(
        n_vertices,
        edge_tuples,
        edge_attrs={
            "length": [attr[0] for attr in edge_attributes],
            "original_id": [attr[1] for attr in edge_attributes],
            "owners": [attr[2] for attr in edge_attributes],
            "edge_type": [EdgeType.RAIL_LINK.value for _ in edge_attributes],
            "origin_coords": [attr[3] for attr in edge_attributes],
            "destination_coords": [attr[4] for attr in edge_attributes],
            "tracks": [attr[5] for attr in edge_attributes],
        },
        directed=True,
    )
    context.log.info(f"Initial constructed graph connected?: {g.is_connected()}")
    allowed_node_indices = get_allowed_node_indices(g)

    # construct a connected subgraph
    connected_subgraph = g.subgraph(allowed_node_indices)
    context.log.info(f"Graph has {len(g.vs)} nodes and {len(g.es)} edges.")
    return connected_subgraph


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def impedance_rail_graph(
    context: dagster.AssetExecutionContext,
    strongly_connected_rail_graph: ig.Graph,
) -> ig.Graph:
    """iGraph object representing the impedance network, derived from the rail network"""
    g = generate_impedance_graph(strongly_connected_rail_graph, SEPARATION_ATTRIBUTE_NAME)
    context.log.info(f"Graph has {len(g.vs)} nodes and {len(g.es)} edges.")
    assert g.is_connected()
    return g


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def impedance_rail_graph_with_terminals(
    context: dagster.AssetExecutionContext,
    intermodal_terminals_src: pd.DataFrame,
    impedance_rail_graph: ig.Graph,
) -> ig.Graph:
    """Add intermodal facilities and attach them to the impedance rail graph at the right place.
    This asset matches owners at the intermodal facilities to the closest impedance graph node(s)"""

    # convert RAIL CO to a set, replacing things inside of parentheses and that occur with large strings ending in '-'
    intermodal_idx_to_rail_carriers = (
        intermodal_terminals_src["RAIL_CO"]
        .str.replace("via", ",")
        .str.replace(" ", "")
        .str.replace(r"\(.*\)", "", regex=True)
        .replace(r".*-", "", regex=True)
        .str.split(",")
        .apply(lambda x: set(x))
    )

    def replace_items(item_set):
        return {RR_MAPPING.get(item, item) for item in item_set}

    intermodal_terminals_src["RAIL_CO"] = intermodal_idx_to_rail_carriers.apply(replace_items)

    intersection_of_im_and_rail = intermodal_terminals_src["RAIL_CO"].apply(
        lambda x: x & set(impedance_rail_graph.es["owners"])
    )
    im_terminals_found_in_rail = intersection_of_im_and_rail.loc[
        intersection_of_im_and_rail != set()
    ]

    context.log.info(
        f"{len(im_terminals_found_in_rail)} IM terminals found in rail network out of {len(intermodal_terminals_src)}"
    )
    im_pdf = intermodal_terminals_src.iloc[im_terminals_found_in_rail.index].reset_index()

    # add vertices for the terminals and dummy nodes
    num_vertices_to_add = len(im_pdf) * 2
    vertex_attributes = defaultdict(list)
    for row in im_pdf.itertuples():
        for vertex_type in [VertexType.IM_TERMINAL, VertexType.IM_DUMMY_NODE]:
            vertex_attributes["terminal_idx"].append(row.Index)
            vertex_attributes["terminal_name"].append(row.TERMINAL)
            vertex_attributes["coords"].append((row.LAT, row.LON))
            vertex_attributes["vertex_type"].append(vertex_type.value)
            vertex_attributes["owners"].append(row.RAIL_CO)

    impedance_rail_graph.add_vertices(num_vertices_to_add, attributes=vertex_attributes)

    # get actual vertex indices for these just-added vertices
    terminal_idx_to_graph_vertex_idx_map = {
        v["terminal_idx"]: v.index
        for v in impedance_rail_graph.vs.select(vertex_type=VertexType.IM_TERMINAL)
    }
    dummy_terminal_idx_to_graph_vertex_idx_map = {
        v["terminal_idx"]: v.index
        for v in impedance_rail_graph.vs.select(vertex_type=VertexType.IM_DUMMY_NODE)
    }

    # create a map of (owner, origin coords)->source vertex index for all non-impedance edges
    owner_coord_to_vertex_idx_map = {
        (e["owners"], e["origin_coords"]): e.source_vertex.index
        for e in impedance_rail_graph.es
        if e["edge_type"] != EdgeType.IMPEDANCE_LINK.value
    }

    quant_node_lat_longs = [
        origin_coords for _, origin_coords in owner_coord_to_vertex_idx_map.keys()
    ]
    quant_lat_longs_radians = np.deg2rad(quant_node_lat_longs)
    quant_node_ball_tree = BallTree(quant_lat_longs_radians, metric="haversine")

    im_lat_longs = im_pdf[["LAT", "LON"]]
    im_fac_lat_longs_radians = np.deg2rad(im_lat_longs)

    # look up the closest 1000 nodes to a given terminal - if we find an operator match, great. if not, we discard
    distances_radians, lookup_to_bt_node_idx = quant_node_ball_tree.query(
        im_fac_lat_longs_radians, k=1000
    )
    edges = []
    edge_attributes: Dict[str, list[Any]] = defaultdict(list)
    for im_idx, record in enumerate(im_pdf.itertuples()):
        rrs_at_im: set = record.RAIL_CO
        dummy_terminal_node_vertex_idx = dummy_terminal_idx_to_graph_vertex_idx_map[im_idx]
        im_terminal_node_vertex_idx = terminal_idx_to_graph_vertex_idx_map[im_idx]

        # add edges between the IM terminal and its dummy node
        edges.append((im_terminal_node_vertex_idx, dummy_terminal_node_vertex_idx))
        edges.append((dummy_terminal_node_vertex_idx, im_terminal_node_vertex_idx))
        for i in range(2):
            edge_attributes["edge_type"].append(EdgeType.IM_CAPACITY.value)
            edge_attributes["length"].append(0.1)  # nominal length

        # now try to figure out where to map to the quant network
        for rr_at_im in rrs_at_im:
            for candidate_quant_node in lookup_to_bt_node_idx[im_idx]:
                # try to get a vertex with the particular owner and the matching lat long from the quant network
                matching_vertex = owner_coord_to_vertex_idx_map.get(
                    (rr_at_im, quant_node_lat_longs[candidate_quant_node])
                )
                if matching_vertex:
                    edges.append(
                        (dummy_terminal_idx_to_graph_vertex_idx_map[im_idx], matching_vertex)
                    )
                    edges.append(
                        (matching_vertex, dummy_terminal_idx_to_graph_vertex_idx_map[im_idx])
                    )
                    for i in range(2):
                        edge_attributes["length"].append(0.1)  # nominal length
                        edge_attributes["edge_type"].append(EdgeType.IM_DUMMY.value)
                    break
            else:
                context.log.info(
                    f"Nothing found for {rr_at_im} for terminal number {im_idx} with name"
                    f" {record.TERMINAL}"
                )

    impedance_rail_graph.add_edges(edges, attributes=edge_attributes)
    context.log.info(
        f"Graph has {len(impedance_rail_graph.vs)} nodes and {len(impedance_rail_graph.es)} edges."
    )
    assert impedance_rail_graph.is_connected()
    return impedance_rail_graph


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def impedance_rail_graph_with_terminals_reduced(
    context: dagster.AssetExecutionContext,
    impedance_rail_graph_with_terminals: ig.Graph,
) -> ig.Graph:
    """Compute IM to IM shortest paths and reduce the rail impedance network to only consider these edges"""
    im_indices = [
        v.index
        for v in impedance_rail_graph_with_terminals.vs.select(
            vertex_type=VertexType.IM_TERMINAL.value
        )
    ]

    # get all the edges for IM->IM on the impedance network
    edge_set: set[int] = set()
    for idx, im_from in enumerate(im_indices):
        destinations = [i for i in im_indices if i != im_from]
        edge_results = impedance_rail_graph_with_terminals.get_shortest_paths(
            im_from, destinations, weights="length", output="epath"
        )
        edge_set.update(chain.from_iterable(edge_results))

    # get the edges between IM->IM_dummy and IM_dummy->Impedance network
    im_capacity_edges = [
        e.index
        for e in impedance_rail_graph_with_terminals.es.select(edge_type=EdgeType.IM_CAPACITY)
    ]
    im_dummy_edges = [
        e.index for e in impedance_rail_graph_with_terminals.es.select(edge_type=EdgeType.IM_DUMMY)
    ]
    edge_set.update(im_capacity_edges)
    edge_set.update(im_dummy_edges)
    small_g = impedance_rail_graph_with_terminals.subgraph_edges(edge_set)
    context.log.info(f"Graph has {len(small_g.vs)} nodes and {len(small_g.es)} edges.")
    assert small_g.is_connected()
    return small_g



from collections import defaultdict
from typing import Dict, Tuple

import dagster
import geopandas
import igraph as ig
import numpy as np
from sklearn.neighbors import BallTree

from ireiat.config.constants import (
    EXCLUDED_FIPS_CODES_MAP,
    LATLONG_CRS,
    ALBERS_CRS,
    RADIUS_EARTH_MILES,
)
from ireiat.config.data_pipeline import TAPRailConfig
from ireiat.config.rail_enum import EdgeType, VertexType


def _generate_network_indices_from_ball_tree(
    context: dagster.AssetExecutionContext,
    county_centroids: Dict[Tuple[str, str], Tuple[float, float]],
    bt: BallTree,
) -> Dict[Tuple[str, str], int]:
    """Helper function for assets"""
    centroid_radians = np.deg2rad(np.array(np.array(list(county_centroids.values()))))
    distances_radians, centroid_idx_to_network_node_idx = bt.query(centroid_radians, k=1)
    distances_miles = distances_radians.squeeze() * RADIUS_EARTH_MILES
    centroid_idx_to_network_node_idx = (
        centroid_idx_to_network_node_idx.squeeze()
    )  # make a 1-D array

    context.log.info(f"Total node indices: {len(centroid_idx_to_network_node_idx)}")
    context.log.info(f"Unique node indices: {len(set(centroid_idx_to_network_node_idx))}")
    context.log.info(
        "If these aren't the same, then (state,county) combos point to the same network node, fyi"
    )
    context.log.info(f"Max distance a node is: {max(distances_miles)} miles")

    return {k: int(v) for k, v in zip(county_centroids.keys(), centroid_idx_to_network_node_idx)}


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def county_fips_to_centroid(
    us_county_shp_files_src: geopandas.GeoDataFrame,
) -> Dict[Tuple[str, str], Tuple[float, float]]:
    """Returns (STATE_FIPS, COUNTY_FIPS) -> (latitude, longitude) based on county centroids"""
    # exclude states that we don't care about
    county_gdf = us_county_shp_files_src.loc[
        ~(us_county_shp_files_src["STATEFP"].isin(EXCLUDED_FIPS_CODES_MAP.values()))
    ]
    county_centroids = county_gdf.to_crs(ALBERS_CRS).geometry.centroid.to_crs(LATLONG_CRS)

    # create mapping: (STATE FIPS, COUNTY FIPS) -> county centroid
    county_centroid_dict: Dict[Tuple[str, str], Tuple[float, float]] = dict()
    for idx, (row, centroid) in enumerate(zip(county_gdf.itertuples(), county_centroids)):
        key = (row.STATEFP, row.COUNTYFP)
        county_centroid_dict[key] = (round(centroid.y, 6), round(centroid.x, 6))

    return county_centroid_dict


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def county_fips_to_highway_network_node_idx(
    context: dagster.AssetExecutionContext,
    county_fips_to_centroid: Dict[Tuple[str, str], Tuple[float, float]],
    highway_ball_tree: BallTree,
) -> Dict[Tuple[str, str], int]:
    """Map all county centroids to the nearest highway nodes returning a dict of (STATE, COUNTY) -> highway node"""
    return _generate_network_indices_from_ball_tree(
        context, county_fips_to_centroid, highway_ball_tree
    )


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def county_fips_to_marine_network_node_idx(
    context: dagster.AssetExecutionContext,
    county_fips_to_centroid: Dict[Tuple[str, str], Tuple[float, float]],
    marine_ball_tree: BallTree,
) -> Dict[Tuple[str, str], int]:
    """Map all county centroids to the nearest marine nodes returning a dict of (STATE, COUNTY) -> marine node"""
    return _generate_network_indices_from_ball_tree(
        context, county_fips_to_centroid, marine_ball_tree
    )


@dagster.multi_asset(
    outs={
        "rail_graph_with_county_connections": dagster.AssetOut(
            io_manager_key="default_io_manager_intermediate_path"
        ),
        "county_fips_to_rail_network_node_idx": dagster.AssetOut(
            io_manager_key="default_io_manager_intermediate_path"
        ),
    }
)
def rail_county_association(
    context: dagster.AssetExecutionContext,
    impedance_rail_graph_with_terminals_reduced: ig.Graph,
    county_fips_to_centroid: Dict[Tuple[str, str], Tuple[float, float]],
    config: TAPRailConfig,
) -> Tuple[ig.Graph, Dict[Tuple[str, str], int]]:
    """Add county centroids to rail impedance network with county centroids connected to nearby IM facilities."""

    # unfortunately, there are several counties that may not have ANY intermodal facilities within a radius
    # we don't want to add these counties as vertices to the graph, so we need to filter them out first
    # by interrogating which IM facilities are nearby ALL counties
    g = impedance_rail_graph_with_terminals_reduced
    im_lat_longs = [v["coords"] for v in g.vs.select(vertex_type=VertexType.IM_TERMINAL.value)]
    im_lat_longs_radians = np.deg2rad(np.array(im_lat_longs))
    im_node_ball_tree = BallTree(im_lat_longs_radians, metric="haversine")

    search_radius_radians = config.intermodal_search_radius_miles / RADIUS_EARTH_MILES
    county_lat_longs = list(county_fips_to_centroid.values())
    reverse_county_lookup = {v: k for k, v in county_fips_to_centroid.items()}
    search_coords_radians = np.deg2rad(np.array(county_lat_longs))
    imfac_idxs, distances_radians = im_node_ball_tree.query_radius(
        search_coords_radians, r=search_radius_radians, return_distance=True, sort_results=True
    )
    county_fips_to_centroid_with_im = dict()
    for idx, candidate_ims in enumerate(imfac_idxs):
        if len(candidate_ims) > 0:
            key = reverse_county_lookup[county_lat_longs[idx]]
            county_fips_to_centroid_with_im[key] = county_fips_to_centroid[key]

    # log the results
    counties_without_im = len(county_fips_to_centroid) - len(county_fips_to_centroid_with_im)
    context.log.info(
        f"{counties_without_im} counties do not have IM facilities nearby. Excluding them from the graph."
    )

    # now we have a list of county centroids with IM facilities, which we add to the graph
    attr_dict = {
        "coords": [v for v in county_fips_to_centroid_with_im.values()],
        "vertex_type": [VertexType.COUNTY_CENTROID for _ in county_fips_to_centroid_with_im],
    }
    g.add_vertices(len(county_fips_to_centroid_with_im), attributes=attr_dict)

    # look up the coordinates and of each centroid vertex and map to the vertex ID (after adding centroid vertices)
    county_centroid_coords_to_vertex_idx = {
        v["coords"]: v.index for v in g.vs.select(vertex_type=VertexType.COUNTY_CENTROID.value)
    }

    # look up the coordinates and of each IM vertex and map to the vertex ID (after adding centroid vertices)
    im_coords_to_vertex_idx = {
        v["coords"]: v.index for v in g.vs.select(vertex_type=VertexType.IM_TERMINAL.value)
    }

    # create a ball tree
    im_node_lat_longs = list(im_coords_to_vertex_idx.keys())
    im_lat_longs_radians = np.deg2rad(np.array(im_node_lat_longs))
    im_node_ball_tree = BallTree(im_lat_longs_radians, metric="haversine")

    search_radius_radians = config.intermodal_search_radius_miles / RADIUS_EARTH_MILES
    county_lat_longs = list(county_centroid_coords_to_vertex_idx.keys())
    search_coords_radians = np.deg2rad(np.array(county_lat_longs))
    imfac_idxs, distances_radians = im_node_ball_tree.query_radius(
        search_coords_radians, r=search_radius_radians, return_distance=True, sort_results=True
    )
    distances_miles = distances_radians * RADIUS_EARTH_MILES

    edges_to_add = []
    edge_attributes = defaultdict(list)
    for idx, (candidate_ims, distances) in enumerate(zip(imfac_idxs, distances_miles)):
        for candidate_im, distance in zip(candidate_ims, distances):
            im_vertex = im_coords_to_vertex_idx[im_node_lat_longs[candidate_im]]
            county_vertex = county_centroid_coords_to_vertex_idx[county_lat_longs[idx]]
            edges_to_add.append((county_vertex, im_vertex))
            edges_to_add.append((im_vertex, county_vertex))
            for _ in range(2):
                edge_attributes["length"].append(distance)
                edge_attributes["edge_type"].append(EdgeType.COUNTY_TO_IM_LINK.value)
                edge_attributes["speed"].append(
                    config.dray_default_speed_mph / config.dray_penalty_factor
                )

    g.add_edges(edges_to_add, attributes=edge_attributes)

    assert g.is_connected()

    county_fips_to_rail_network_node_idx = {
        k: county_centroid_coords_to_vertex_idx[v]
        for k, v in county_fips_to_centroid_with_im.items()
    }
    return g, county_fips_to_rail_network_node_idx



import dagster
from ireiat.data_pipeline.io_manager import asset_spec_factory

marine_links_spec = dagster.AssetSpec(
    key=dagster.AssetKey("marine_network_links_spec"),
    description="Publicly available US Army Corps of Engineers data for marine network links",
    metadata={
        "format": "zip",
        "filename": "marine_links.zip",
        "source_path": "raw/",
        "dashboard_url": dagster.MetadataValue.url(
            "https://usace.contentdm.oclc.org/digital/api/collection/p16021coll2/id/3010/download"
        ),
    },
)
marine_links_src = asset_spec_factory(marine_links_spec)



import unittest

import igraph as ig

from ireiat.data_pipeline.assets.rail_network.impedance import (
    generate_impedance_graph,
    _generate_subgraphs,
)


class TestImpedances(unittest.TestCase):

    def test_generate_subgraphs_works_for_a_simple_case(self):
        g = ig.Graph(directed=True)
        g.add_vertices(3)
        g.add_edges([(0, 1), (1, 2)])
        g.es["origin_coords"] = [(0, 0), (1, 1)]
        g.es["destination_coords"] = [(1, 1), (2, 2)]
        g.es["owner"] = [{"A", "B"}, {"B"}]
        s = _generate_subgraphs(g)
        self.assertEqual(len(s.keys()), 2)  # just A and B
        a_graph, b_graph = s["A"], s["B"]
        self.assertEqual(len(a_graph.es), 1)  # A graph has 1 edge
        self.assertEqual(len(b_graph.es), 2)  # B graph has 2 edges

    def _node_edge_impedance_confirmation(
        self, g: ig.Graph, expected_nodes: int, expected_edges: int, expected_impedance_edges: int
    ):
        """Helper method used within this test suite"""
        i = generate_impedance_graph(g)
        self.assertEqual(len(i.vs), expected_nodes)  # should have expected nodes
        self.assertEqual(len(i.es), expected_edges)  # should have expected edges
        self.assertEqual(
            len(i.es.select(owner_eq="imp")), expected_impedance_edges
        )  # should have expected impedance edge, relies on 'owner' being the separation attribute

    def test_simple_graph_establishes_sensible_impedances(self):
        g = ig.Graph(directed=True)
        g.add_vertices(4)
        g.add_edges([(0, 1), (1, 2), (2, 3)])
        g.es["owner"] = [{"A", "B"}, {"A", "B"}, {"B"}]
        g.es["origin_coords"] = [(0, 0), (1, 1), (2, 2)]
        g.es["destination_coords"] = [(1, 1), (2, 2), (3, 3)]
        self._node_edge_impedance_confirmation(g, 7, 6, 1)

    def test_junction_with_two_outlet_owners_establishes_sensible_impedances(self):
        g = ig.Graph(directed=True)
        g.add_vertices(4)
        g.add_edges([(0, 1), (1, 2), (1, 3)])
        g.es["owner"] = [{"A"}, {"A"}, {"C"}]
        g.es["origin_coords"] = [(0, 0), (1, 1), (1, 1)]
        g.es["destination_coords"] = [(1, 1), (2, 2), (3, 3)]
        self._node_edge_impedance_confirmation(g, 5, 4, 1)

    def test_junction_with_two_inlet_and_two_outlet_establishes_impedances(self):
        g = ig.Graph(directed=True)
        g.add_vertices(5)
        g.add_edges([(0, 1), (1, 2), (3, 1), (1, 4)])
        g.es["owner"] = [{"A"}, {"A"}, {"C"}, {"C"}]
        g.es["origin_coords"] = [(0, 0), (1, 1), (3, 3), (1, 1)]
        g.es["destination_coords"] = [(1, 1), (2, 2), (1, 1), (4, 4)]
        self._node_edge_impedance_confirmation(g, 6, 6, 2)

    def test_bi_directional_junction_with_two_inlet_and_two_outlet_establishes_impedances(self):
        g = ig.Graph(directed=True)
        g.add_vertices(5)
        edges = [(0, 1), (1, 0), (1, 2), (2, 1), (3, 1), (1, 3), (1, 4), (4, 1)]
        g.add_edges(edges)
        g.es["owner"] = [{"A"}, {"A"}, {"C"}, {"C"}]
        g.es["origin_coords"] = [(o, o) for o, _ in edges]
        g.es["destination_coords"] = [(d, d) for _, d in edges]
        self._node_edge_impedance_confirmation(g, 6, 10, 2)

    def test_multiple_owners_establishes_impedances(self):
        g = ig.Graph(directed=True)
        g.add_vertices(5)
        edges = [(0, 1), (1, 2), (2, 3), (2, 4)]
        g.add_edges(edges)
        g.es["owner"] = [{"A", "B"}, {"B"}, {"A", "B"}, {"C"}]
        g.es["origin_coords"] = [(o, o) for o, _ in edges]
        g.es["destination_coords"] = [(d, d) for _, d in edges]
        self._node_edge_impedance_confirmation(g, 10, 9, 3)



import dagster

from ireiat.data_pipeline.io_manager import asset_spec_factory

FAF5_REGIONS_DESCRIPTION = "Publicly available GIS data for FAF5 regions"
faf5_regions_spec = dagster.AssetSpec(
    key=dagster.AssetKey("faf5_regions_spec"),
    description=FAF5_REGIONS_DESCRIPTION,
    metadata={
        "format": "zip",
        "filename": "faf5_regions.zip",
        "source_path": "raw/",
        "dashboard_url": dagster.MetadataValue.url(
            "https://hub.arcgis.com/api/download/v1/items/539fdab88ee74e38b28959494965ace2/shapefile?redirect=true&layers=0"
        ),
    },
)

faf5_regions_src = asset_spec_factory(faf5_regions_spec)

US_COUNTY_SHP_FILES_DESCRIPTION = "Publicly available GIS data for US counties"
us_county_shp_files_spec = dagster.AssetSpec(
    key=dagster.AssetKey("us_county_shp_files_spec"),
    description=US_COUNTY_SHP_FILES_DESCRIPTION,
    metadata={
        "format": "zip",
        "filename": "us_county_shp_files.zip",
        "source_path": "raw/",
        "dashboard_url": dagster.MetadataValue.url(
            "https://www2.census.gov/geo/tiger/TIGER2023/COUNTY/tl_2023_us_county.zip"
        ),
    },
)
us_county_shp_files_src = asset_spec_factory(us_county_shp_files_spec)

US_CENSUS_COUNTY_POPULATION_DESCRIPTION = (
    "Publicly available census population data for US counties"
)
us_census_county_population_spec = dagster.AssetSpec(
    key=dagster.AssetKey("us_census_county_population_spec"),
    description=US_CENSUS_COUNTY_POPULATION_DESCRIPTION,
    metadata={
        "format": "csv",
        "filename": "co-est2022-alldata.csv",
        "source_path": "raw/",
        "read_kwargs": dagster.MetadataValue.json({"dtype": {"STATE": "str", "COUNTY": "str"}}),
        "dashboard_url": dagster.MetadataValue.url(
            "https://www2.census.gov/programs-surveys/popest/datasets/2020-2022/counties/totals/co-est2022-alldata.csv"
        ),
    },
)
us_census_county_population_src = asset_spec_factory(us_census_county_population_spec)

FAF_DEMAND_DESCRIPTION = "FAF5 Framework tonnage, ton miles, and values"
faf5_demand_spec = dagster.AssetSpec(
    key=dagster.AssetKey("faf5_demand_spec"),
    description=FAF_DEMAND_DESCRIPTION,
    metadata={
        "format": "zip",
        "filename": "faf5_demand.zip",
        "temp_unzip": True,
        "source_path": "raw/",
        "read_kwargs": dagster.MetadataValue.json(
            {"dtype": {"dms_orig": "str", "dms_dest": "str", "sctg2": "str"}}
        ),
        "dashboard_url": dagster.MetadataValue.url(
            "https://faf.ornl.gov/faf5/data/download_files/FAF5.5.1.zip"
        ),
    },
)
faf5_demand_src = asset_spec_factory(faf5_demand_spec)



import dagster

from .assets import demand, highway_network, tap, rail_network, marine_network
from .io_manager import TabularDataLocalIOManager
from ireiat.config.constants import CACHE_PATH, INTERMEDIATE_PATH

# demand
demand_assets = dagster.load_assets_from_package_module(demand, group_name="demand")
demand_assets_job = dagster.define_asset_job(name="demand_job")

# highway network
highway_network_assets = dagster.load_assets_from_package_module(
    highway_network, group_name="highway_network"
)
highway_network_assets_job = dagster.define_asset_job(name="highway_network_job")

# rail network
rail_network_assets = dagster.load_assets_from_package_module(
    rail_network, group_name="rail_network"
)
rail_network_assets_job = dagster.define_asset_job(
    name="rail_network_job",
)

# marine network
marine_network_assets = dagster.load_assets_from_package_module(
    marine_network, group_name="marine_network"
)
marine_network_assets_job = dagster.define_asset_job(name="marine_network_job")

# tap assets
tap_assets = dagster.load_assets_from_package_module(tap, group_name="tap")
tap_assets_job = dagster.define_asset_job(name="tap_job")

# all assets
all_assets = [
    *demand_assets,
    *highway_network_assets,
    *rail_network_assets,
    *marine_network_assets,
    *tap_assets,
]
all_assets_job = dagster.define_asset_job(name="all")

intermediate_path = str(CACHE_PATH / INTERMEDIATE_PATH)
defs = dagster.Definitions(
    assets=all_assets,
    jobs=[
        demand_assets_job,
        highway_network_assets_job,
        rail_network_assets_job,
        marine_network_assets_job,
        tap_assets_job,
        all_assets_job,
    ],
    resources={
        "default_io_manager": dagster.FilesystemIOManager(base_dir=str(CACHE_PATH)),
        "default_io_manager_intermediate_path": dagster.FilesystemIOManager(
            base_dir=intermediate_path
        ),
        "custom_io_manager": TabularDataLocalIOManager(),
    },
)



import unittest

import igraph
import pandas as pd

from ireiat.util.graph import generate_zero_based_node_maps, get_allowed_node_indices


class TestGIS(unittest.TestCase):

    def setUp(self) -> None:
        cols = [
            "origin_latitude",
            "origin_longitude",
            "destination_latitude",
            "destination_longitude",
        ]
        record_with_duplicate_origins = [(1, 1, 2, 2), (1, 1, 3, 3)]
        self.duplicate_origin_df = pd.DataFrame(record_with_duplicate_origins, columns=cols)
        record_with_duplicate_dest = [(1, 1, 2, 2), (3, 3, 2, 2)]
        self.duplicate_destination_df = pd.DataFrame(record_with_duplicate_dest, columns=cols)

    def test_generate_zero_nodes_with_duplicate_origin(self):
        dict_result = generate_zero_based_node_maps(self.duplicate_origin_df)
        self.assertEqual(len(dict_result), 3)
        self.assertEqual(max([v for v in dict_result.values()]), 2)  # max index = count-1

    def test_generate_zero_nodes_with_duplicate_destination(self):
        dict_result = generate_zero_based_node_maps(self.duplicate_destination_df)
        self.assertEqual(len(dict_result), 3)

    def test_fully_connected_graph_returns_all_node_indices(self):
        g = igraph.Graph(edges=[(1, 2), (2, 3)])
        result = get_allowed_node_indices(g)
        self.assertEqual(len(result), 3)  # nodes 1,2,3 all connected

    def test_partly_connected_graph_returns_biggest_connected_component(self):
        g = igraph.Graph(edges=[(1, 2), (2, 3), (5, 6)])
        result = get_allowed_node_indices(g)
        self.assertEqual(len(result), 3)  # nodes 1,2,3 all connected - nodes 5,6 not connected



import dagster
import geopandas
import pandas as pd


def publish_metadata(
    context: dagster.AssetExecutionContext,
    pdf: pd.DataFrame | geopandas.GeoDataFrame,
    output_name: str = None,
) -> None:
    """Publishes metadata for pandas dataframes given a dagster execution context.
    Note excludes any column named "geometry" since these GIS data make reading tabular data
    difficult."""
    temp_df = pdf[[c for c in pdf.columns if c != "geometry"]]

    metadata = {
        "rows": len(temp_df),
        "preview": dagster.MetadataValue.md(temp_df.head().to_markdown()),
    }

    # If output_name is provided, use it; otherwise, log metadata without it
    if output_name:
        context.add_output_metadata(metadata, output_name=output_name)
    else:
        context.add_output_metadata(metadata)



from typing import Optional

import yaml
from dagster import Config
from pydantic import Field

FAF5_DEFAULT_TONS_FIELD = "tons_2022"


class FAFCommodity(Config):
    """Single FAF Commodity"""

    name: str
    sctg2: str = Field(min_length=2, max_length=2)
    containerizable: bool = Field(default=False)
    percentage_containerizable: float = Field(default=0.0, le=1, ge=0)

    @classmethod
    def from_args(cls, name, sctg2, containerizable, pct):
        arg_dict = {
            "name": name,
            "sctg2": sctg2,
            "containerizable": containerizable,
            "percentage_containerizable": pct,
        }
        return cls(**arg_dict)


def default_faf_commodity_factory() -> list[FAFCommodity]:
    return [
        FAFCommodity.from_args("LIVE_ANIMALS_FISH", "01", False, 0.0),
        FAFCommodity.from_args("CEREAL_GRAINS", "02", False, 0.0),
        FAFCommodity.from_args("OTHER_AG_PRODS", "03", False, 0.0),
        FAFCommodity.from_args("ANIMAL_FEED", "04", False, 0.0),
        FAFCommodity.from_args("MEAT_SEAFOOD", "05", True, 0.159),
        FAFCommodity.from_args("MILLED_GRAIN_PRODS", "06", True, 0.018),
        FAFCommodity.from_args("OTHER_FOODSTUFFS", "07", True, 0.504),
        FAFCommodity.from_args("ALCOHOLIC_BEVERAGES", "08", True, 0.455),
        FAFCommodity.from_args("TOBACCO_PRODS", "09", True, 1.0),
        FAFCommodity.from_args("BUILDING_STONE", "10", False, 0.0),
        FAFCommodity.from_args("NATURAL_SANDS", "11", False, 0.0),
        FAFCommodity.from_args("GRAVEL", "12", False, 0.0),
        FAFCommodity.from_args("NONMETALLIC_MINERALS", "13", False, 0.0),
        FAFCommodity.from_args("METALLIC_ORES", "14", False, 0.0),
        FAFCommodity.from_args("COAL", "15", False, 0.0),
        FAFCommodity.from_args("CRUDE_PETROLEUM", "16", False, 0.0),
        FAFCommodity.from_args("GASOLINE", "17", False, 0.0),
        FAFCommodity.from_args("FUEL_OILS", "18", False, 0.0),
        FAFCommodity.from_args("NATURAL_GAS_AND_OTHER_FOSSIL_PRODUCTS", "19", False, 0.0),
        FAFCommodity.from_args("BASIC_CHEMICALS", "20", False, 0.0),
        FAFCommodity.from_args("PHARMACEUTICALS", "21", True, 0.578),
        FAFCommodity.from_args("FERTILIZERS", "22", False, 0.0),
        FAFCommodity.from_args("CHEMICAL_PRODS", "23", True, 0.016),
        FAFCommodity.from_args("PLASTICS_RUBBER", "24", True, 0.056),
        FAFCommodity.from_args("LOGS", "25", False, 0.0),
        FAFCommodity.from_args("WOOD_PRODS", "26", False, 0.0),
        FAFCommodity.from_args("NEWSPRINT_PAPER", "27", False, 0.0),
        FAFCommodity.from_args("PAPER_ARTICLES", "28", True, 0.998),
        FAFCommodity.from_args("PRINTED_PRODS", "29", True, 1.0),
        FAFCommodity.from_args("TEXTILES_LEATHER", "30", True, 0.02),
        FAFCommodity.from_args("NONMETAL_MIN_PRODS", "31", False, 0.0),
        FAFCommodity.from_args("BASE_METALS", "32", False, 0.0),
        FAFCommodity.from_args("ARTICLESBASE_METAL", "33", True, 0.21),
        FAFCommodity.from_args("MACHINERY", "34", True, 0.593),
        FAFCommodity.from_args("ELECTRONICS", "35", True, 0.56),
        FAFCommodity.from_args("MOTORIZED_VEHICLES", "36", True, 0.641),
        FAFCommodity.from_args("TRANSPORT_EQUIP", "37", True, 0.031),
        FAFCommodity.from_args("PRECISION_INSTRUMENTS", "38", True, 0.021),
        FAFCommodity.from_args("FURNITURE", "39", True, 0.05),
        FAFCommodity.from_args("MISC_MFG_PRODS", "40", True, 0.108),
        FAFCommodity.from_args("WASTE_SCRAP", "41", False, 0.0),
        FAFCommodity.from_args("MIXED_FREIGHT", "43", True, 0.988),
    ]


class FAF5MasterConfig(Config):
    """Parent class to allow us to inject the FAF demand field into child classes"""

    faf_demand_field: str = Field(
        default=FAF5_DEFAULT_TONS_FIELD,
        description="The field in FAF used to aggregate demand ( in tons)",
    )


class FAF5FilterConfig(FAF5MasterConfig):
    """Allocates whether FAF commodities are considered as part of the demand. See FAF SCTG2 codes.
    https://www.bts.gov/sites/bts.dot.gov/files/2021-02/FAF5-User-Guide.pdf"""

    faf_commodities: list[FAFCommodity] = Field(default_factory=default_faf_commodity_factory)


class FAF5DemandConfig(FAF5MasterConfig):
    """Used to group and allocate unknown modes to this mode"""

    unknown_mode_percent: float = Field(
        default=0.3,
        description="Percentage of mixed mode FAF demand to allocate to the given modal network",
    )


class TAPFilterTonsConfig(Config):
    """Used to specify an optional quantile threshold to filter county|county tons for the mode"""

    quantile_threshold: Optional[float] = Field(
        default=0.99,
        description="If specified, only includes a county|county highway tonnage above the quantile value",
    )


class TAPNetworkConfig(Config):
    """Used to specify default configuration parameters"""

    default_speed_mph: int = Field(
        default=20,
        description="Default speed to use for (non-dray) rail links in the rail TAP network",
    )
    default_capacity_ktons: int = Field(
        default=100_000,
        description="Default capacity (in kilotons) to use for network edges without a custom capacity",
    )
    default_network_alpha: float = Field(
        default=0.1,
        description="Scalar, alpha, in TAP congestion term (alpha * (flow/capacity) ^ beta)",
    )
    default_network_beta: float = Field(
        default=2.0,
        description="Exponent, beta, in TAP congestion term (alpha * (flow/capacity) ^ beta)",
    )


class TAPRailConfig(TAPNetworkConfig):
    """Adjustable configuration parameters for running the rail network"""

    dray_default_speed_mph: int = Field(
        default=50, description="Default speed for drayage legs (for road segments) in the rail TAP"
    )
    dray_penalty_factor: int = Field(
        default=100,
        description="Factor to increase FFT on drayage legs to avoid exclusive use of the road network for the "
        "rail TAP",
    )
    capacity_ktons_per_track: int = Field(
        default=1000,
        description="Per track capacity (in kilotons) used in the rail network for NARN edges",
    )
    capacity_ktons_default: int = Field(
        default=100_000,
        description="Default capacity (in kilotons) to use for non-IM capacity and non-rail network (i.e. dray, "
        "impedance, etc.)",
    )
    intermodal_search_radius_miles: int = Field(
        default=250,
        description="Distance to search from an origin/destination centroid to an intermodal facility",
    )
    intermodal_facility_capacity_ktons: int = Field(
        default=7500,
        description="Default capacity (in kiltons) to use for IM facility inlet/outlet edges",
    )
    intermodal_edge_alpha: float = Field(
        default=1,
        description="Scalar, alpha, in TAP congestion term (alpha * (flow/capacity) ^ beta) for IM capacity links",
    )
    intermodal_edge_beta: float = Field(
        default=1,
        description="Exponent, beta, in TAP congestion term (alpha * (flow/capacity) ^ beta) for IM capacity links",
    )


default_tap_marine_config = TAPNetworkConfig(
    **{
        "default_speed_mph": 15,
        "default_capacity_ktons": 60_000,
        "default_network_alpha": 0.1,
        "default_network_beta": 2,
    }
)

default_tap_highway_config = TAPNetworkConfig(
    **{
        "default_speed_mph": 50,
        "default_capacity_ktons": 100_000,
        "default_network_alpha": 0.15,
        "default_network_beta": 4,
    }
)

default_tap_rail_config = TAPRailConfig(
    **{
        "default_speed_mph": 20,
        "default_capacity_ktons": 100_000,
        "default_network_alpha": 0.1,
        "default_network_beta": 2.0,
    }
)


def default_asset_mapping() -> dict:
    return {
        "faf_filtered_grouped_tons": {"config": FAF5FilterConfig()},
        "faf5_truck_demand": {"config": FAF5DemandConfig(**{"unknown_mode_percent": 0.3})},
        "faf5_rail_demand": {"config": FAF5DemandConfig(**{"unknown_mode_percent": 0.5})},
        "faf5_water_demand": {"config": FAF5DemandConfig(**{"unknown_mode_percent": 0.2})},
        "county_to_county_highway_tons": {"config": FAF5MasterConfig()},
        "county_to_county_rail_tons": {"config": FAF5MasterConfig()},
        "county_to_county_marine_tons": {"config": FAF5MasterConfig()},
        "tap_highway_tons": {"config": TAPFilterTonsConfig(**{"quantile_threshold": 0.999})},
        "tap_rail_tons": {"config": TAPFilterTonsConfig(**{"quantile_threshold": 0.9})},
        "tap_marine_tons": {"config": TAPFilterTonsConfig(**{"quantile_threshold": 0.6})},
        "tap_highway_network_dataframe": {"config": default_tap_highway_config},
        "tap_rail_network_dataframe": {"config": default_tap_rail_config},
        "tap_marine_network_dataframe": {"config": default_tap_marine_config},
    }


class DataPipelineConfig(Config):
    """Follows Dagster's specific configuration requirements for asset-specific configs"""

    ops: dict = Field(
        default_factory=default_asset_mapping,
        description="List of assets and associated configuration",
    )


# only used to update the root config file from the IDE when doing local dev (providing an example file)
if __name__ == "__main__":
    with open("../../../data/config.yaml", "w") as fp:
        yaml.dump(DataPipelineConfig().dict(), fp)



import os
from pathlib import Path

# cache-related parameters
CACHE_PATH = Path(os.getenv("HOMEPATH", "~")) / ".ireiat"
INTERMEDIATE_PATH = "intermediate"
RAW_PATH = "raw"
INTERMEDIATE_DIRECTORY_ARGS = {"source_path": INTERMEDIATE_PATH}

# GIS-related parameters
RADIUS_EARTH_MILES = 3958.8
METERS_PER_MILE = 1609.34
LATLONG_CRS = "EPSG:4326"
ALBERS_CRS = "EPSG:5070"

# FAF-related parameters
SUM_TONS_TOLERANCE = 1e-5

#: Exclude some states, regions from the input data
EXCLUDED_FIPS_CODES_MAP = {
    "Alaska": "02",
    "Hawaii": "15",
    "Puerto Rico": "72",
    "American Samoa": "60",
    "Northern Mariana Islands": "69",
    "Guam": "66",
    "US Virgin Islands": "78",
}

# Rail
# standardize railway codes both in the NARN data and on intermodal terminals
RR_MAPPING = {
    "CP": "CPKC",
    "CSX": "CSXT",
    "CPRS": "CPKC",
    "KCS": "CPKC",
    "KCSM": "CPKC",
    "JXPT": "JXP",
}



import tempfile
from functools import partial
from pathlib import Path
from typing import Optional, Mapping, Callable
from zipfile import ZipFile

import dagster
import geopandas
import pandas as pd
import pyogrio

from ireiat.config.constants import CACHE_PATH
from ireiat.data_pipeline.metadata import publish_metadata
from ireiat.util.http import download_uncached_file


def _get_read_function(format: str, metadata: dict = None) -> Callable:
    """Returns a function for file-reading based on the `format` and possibly `metadata`"""
    format_mapping = {
        "csv": pd.read_csv,
        "parquet": (
            geopandas.read_parquet
            if metadata and metadata.get("use_geopandas", False)
            else pd.read_parquet
        ),
        "zip": partial(pyogrio.read_dataframe, use_arrow=True),
        "txt": pd.read_table,
        "xlsx": pd.read_excel,
    }
    if metadata and metadata.get("temp_unzip", False):
        return _temp_unzip_and_read
    if format in format_mapping:
        print(format_mapping)
        return format_mapping[format]
    else:
        raise NotImplementedError(f"Cannot read file of type {format}!")


def _temp_unzip_and_read(fpath, **kwargs):
    """Unzips the zipfile passed at fpath and reads the first csv into a pandas dataframe"""
    zip = ZipFile(fpath)
    with tempfile.TemporaryDirectory() as tmpdirname:
        zip.extractall(path=tmpdirname)
        target_csvs = [f for f in zip.namelist() if f.endswith("csv")]
        pdf = pd.read_csv(zip.open(target_csvs[0]), **kwargs)
        zip.close()
    return pdf


def _get_fs_path(asset_key: dagster.AssetKey, metadata: Optional[Mapping]) -> str:
    """Gets the filesystem path based on the asset key and/or metadata.
    If the full source_path `some_file.csv` is passed, then use that. Otherwise use
    the filename and the default path, which is the user directory
    """
    source_path = CACHE_PATH / metadata.get("source_path", "")
    fname = metadata.get("filename") or f"{asset_key.path[0]}.{metadata.get('format')}"
    read_path = Path(source_path) / fname
    return str(read_path.absolute())


def read_or_attempt_download(
    asset_key: dagster.AssetKey, current_asset_metadata
) -> pd.DataFrame | geopandas.GeoDataFrame:
    """Reads the file given the metadata. If the file does not exist, attempts to download based on url information
    within the metadata. If download info does not exist and the file is not in the filesystem, throws an IOError
    """
    fpath = _get_fs_path(asset_key, current_asset_metadata)

    # if URL specified, attempt download...
    url: Optional[dagster.UrlMetadataValue] = current_asset_metadata.get("dashboard_url")
    if url:
        download_uncached_file(url.url, fpath)

    if not Path(fpath).exists():
        raise IOError(f"No metadata url specified for {fpath}!")

    read_kwargs: Optional[dagster.JsonMetadataValue] = current_asset_metadata.get("read_kwargs")
    parsed_read_kwargs: dict = read_kwargs.data if read_kwargs else dict()

    # figure out how to read this based on the ending
    fmt = fpath.split(".")[-1]
    read_func = _get_read_function(fmt, current_asset_metadata)

    return read_func(fpath, **parsed_read_kwargs)


class TabularDataLocalIOManager(dagster.ConfigurableIOManager):
    """Translates tabular data (csv, txt, xlsx, and shp files) on the local filesystem."""

    def handle_output(self, context, obj: pd.DataFrame | geopandas.GeoDataFrame) -> None:
        """This saves the dataframe according to the format implemented in FileSerializationResolver."""

        fpath = _get_fs_path(context.asset_key, context.metadata)
        Path(fpath).parent.mkdir(exist_ok=True)

        write_kwargs: Optional[dagster.JsonMetadataValue] = context.metadata.get("write_kwargs")
        parsed_write_kwargs: dict = write_kwargs.data if write_kwargs else dict()
        fmt = fpath.split(".")[-1]
        if fmt == "parquet":
            obj.to_parquet(fpath, **parsed_write_kwargs)
        elif fmt == "csv":
            obj.to_csv(fpath, **parsed_write_kwargs)
        elif fmt == "xlsx":
            obj.to_excel(fpath, **parsed_write_kwargs)
        elif fmt == "zip":
            context.log.info("Assuming zip file from download. Skipping persistence.")
        else:
            raise NotImplementedError(f"Cannot write file of type {fmt}!")

    def load_input(self, context) -> pd.DataFrame | geopandas.GeoDataFrame:
        """This reads a dataframe based on file ending and metadata"""
        return read_or_attempt_download(
            context.upstream_output.asset_key, context.upstream_output.metadata
        )


def asset_spec_factory(spec: dagster.AssetSpec):
    """Creates a materialized asset from an asset spec, including potentially downloading it.
    The method relies on a naming convention for asset specs that end in "_spec" and generates
    assets that end in "_src"."""

    @dagster.asset(
        key=spec.key.to_user_string().replace("_spec", "_src"),
        io_manager_key="custom_io_manager",
        metadata=spec.metadata,
        description=spec.description,
    )
    def _asset(context: dagster.AssetExecutionContext):
        result = read_or_attempt_download(spec.key, spec.metadata)

        # TODO: Make adjustments downstream to expect lower case headers, then uncomment below:
        # Convert all column names to lowercase to avoid case-sensitivity issues
        # result = result.rename(columns=lambda x: x.lower())

        publish_metadata(context, result)
        return result

    return _asset



import dagster

from ireiat.data_pipeline.io_manager import asset_spec_factory

faf5_regions_spec = dagster.AssetSpec(
    key=dagster.AssetKey("faf5_highway_network_links_spec"),
    description="Publicly available GIS data for FAF5 highway network links",
    metadata={
        "format": "zip",
        "filename": "faf5_highway_links.zip",
        "source_path": "raw/",
        "dashboard_url": dagster.MetadataValue.url(
            "https://hub.arcgis.com/api/download/v1/items/b33d6a096bb94ca6b5cf5d96bf68a682/shapefile?redirect=true&layers=0"
        ),
    },
)
faf5_regions_src = asset_spec_factory(faf5_regions_spec)



import logging
from datetime import datetime
from pathlib import Path
from typing import Optional


def configure_logging(
    str_format: str = "{asctime} {levelname} ({module}:{lineno:d}) {message}",
    output_file: bool = False,
    output_file_name: Optional[str] = None,
    output_file_dir: Path = Path("logs"),
    level=logging.INFO,
) -> None:
    """
    Configures logging called *once* at the application entry.

    :param str_format: logger string format
    :param output_file: bool, whether to log to an output file in output file directory
    :param output_file_name: optional, log file name (including extension)
        default YMDHms.log
    :param output_file_dir: optional, log output directory
    :param level: optional, logging level
    :return: None
    """
    # get the root logger and set its level
    root_logger = logging.getLogger()
    root_logger.setLevel(level)

    # set logging format
    formatter = logging.Formatter(str_format, style="{", datefmt="%Y-%m-%d %H:%M:%S")

    # add a stream handler
    chandler = logging.StreamHandler()
    chandler.setFormatter(formatter)
    chandler.setLevel(level)
    root_logger.addHandler(chandler)

    # add a file handler if output_file is specified
    if output_file:
        output_file_dir.mkdir(exist_ok=True)
        default_log_key = datetime.now().strftime("%Y%m%d%H%M%S")
        default_log_fname = f"{default_log_key}.log"
        fname = output_file_name if output_file_name else default_log_fname
        fhandler = logging.FileHandler(output_file_dir / fname)
        fhandler.setFormatter(formatter)
        fhandler.setLevel(level)
        root_logger.addHandler(fhandler)



import logging
import os
import shutil
import subprocess
from importlib import resources
from pathlib import Path
from typing import Optional

import click

from ireiat import r_source
from ireiat.config.constants import CACHE_PATH
from ireiat.config.runtime import (
    run_config_map,
    RunConfig,
    PostprocessConfig,
    postprocess_config_map,
)
from ireiat.postprocessing.postprocessor import PostProcessor
from ireiat.util.logging_ import configure_logging

configure_logging(output_file=True)
logger = logging.getLogger(__name__)

MODE_CHOICES = click.Choice(["highway", "marine", "rail"])


@click.group()
@click.option("--debug/--no-debug", default=False)
def cli(debug):
    logger.info(f"Debug mode is {'on' if debug else 'off'}")
    os.makedirs(CACHE_PATH, exist_ok=True)


@cli.command()
def clear_cache():
    """Clears the local cache directory"""
    result: str = input("Are you sure you want to clear the cache? [y/n] ")
    if result.strip().lower() == "y":
        logger.info(f"Deleting files at {CACHE_PATH}")
        shutil.rmtree(CACHE_PATH)


@cli.command()
@click.option(
    "--network-file",
    "-n",
    type=click.Path(exists=True),
)
@click.option(
    "--od-file",
    "-d",
    type=click.Path(exists=True),
    help="A parquet file that represents the demand to be used for the TAP",
)
@click.option(
    "--output-file",
    "-o",
    type=click.Path(),
    help="A parquet file to which the TAP results will be written",
)
@click.option(
    "--mode",
    "-m",
    type=MODE_CHOICES,
    help="If specified, uses defaults file outputs for the given mode unless other parameters are passed",
)
@click.option(
    "--max-gap", "-g", type=float, default=1e-8, help="The relative gap used in Algorithm B"
)
@click.option("--max-iterations", "-i", type=int, default=1, help="Max iterations for Algorithm B")
def solve(
    network_file: Optional[Path],
    od_file: Optional[Path],
    output_file: Optional[Path],
    mode: Optional[str],
    max_gap: float,
    max_iterations: int,
):
    """Runs the TAP solution in R using cppRouting"""

    config = run_config_map.get(mode, RunConfig)(
        passed_network_file_path=network_file,
        passed_od_file_path=od_file,
        passed_output_file_path=output_file,
    )

    # use the bundled 'tap.r' file as a "resource" and create a temporary file to be run by RScript
    temporary_file_path = CACHE_PATH / "local_tap.r"
    with open(temporary_file_path, "w") as tf:
        with resources.open_text(r_source, "tap.r") as r_text:
            tf.write(r_text.read())

    # pass the command for the RScript file
    cmd = [
        "Rscript",
        tf.name,
        config.network_file_path,
        config.od_file_path,
        str(max_gap),
        config.output_file_path,
        str(max_iterations),
    ]
    logger.info(f"Calling subprocess {cmd}")
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
    # Read the output line by line as it becomes available
    for line in process.stdout:
        print(line, end="")
    process.wait()

    temporary_file_path.unlink(missing_ok=True)


@cli.command()
@click.option(
    "--solution",
    "-s",
    type=click.Path(exists=True),
    help="Solution file representing assigned traffic",
)
@click.option(
    "--solution-graph",
    "-g",
    type=click.Path(exists=True),
    help="The strongly connected (pickled igraph) graph on which the problem has been solved",
)
@click.option(
    "--network-shp",
    "-n",
    type=click.Path(exists=True),
    help="The shp file used to represent the underlying network to enable visualization",
)
@click.option(
    "--mode",
    "-m",
    type=MODE_CHOICES,
    help="If specified, uses defaults file outputs for the given mode unless other parameters are passed",
)
def postprocess(solution: Path, solution_graph: Path, network_shp: Path, mode: str):
    """Post processes results and save in the default configured cache path"""
    logger.info("Running postprocessing.")
    config = postprocess_config_map.get(mode, PostprocessConfig)(
        passed_traffic_path=solution,
        passed_network_graph_path=solution_graph,
        passed_shp_file_path=network_shp,
    )

    pp = PostProcessor(config.traffic_file_path, config.network_graph_path, config.shp_file_path)
    pp.postprocess()


if __name__ == "__main__":
    cli()



import tempfile
import unittest
from pathlib import Path
from unittest.mock import create_autospec

import dagster
import pandas as pd

from ireiat.config.constants import CACHE_PATH
from ireiat.data_pipeline.io_manager import (
    _get_read_function,
    _get_fs_path,
    read_or_attempt_download,
)
from ireiat.util.http import download_uncached_file


class TestIOManager(unittest.TestCase):

    def setUp(self):
        self.dummy_df = pd.DataFrame([1, 2, 3], columns=["dummy"])

    def test_read_func_throws_for_unusual_file_format(self):
        with self.assertRaises(NotImplementedError):
            _get_read_function("weird_file_extension")

    def test_src_path_resolves(self):
        result = _get_fs_path(dagster.AssetKey("dummy"), {"format": "dummy_extension"})
        self.assertIsInstance(result, str)
        self.assertEqual(Path(result), (CACHE_PATH / "dummy.dummy_extension").absolute())

        # specifying filename overrides any file format
        result = _get_fs_path(
            dagster.AssetKey("dummy"), {"filename": "new_dummy.blah", "format": "dummy_extension"}
        )
        self.assertEqual(Path(result), (CACHE_PATH / "new_dummy.blah").absolute())

    def test_file_read_from_filesystem_when_file_exists(self):
        temp_fname = "test"
        temp_extension = "csv"
        with tempfile.TemporaryDirectory() as td:
            temp_dir_name = Path(td).parent
            self.dummy_df.to_csv(temp_dir_name / f"{temp_fname}.{temp_extension}")
            result = read_or_attempt_download(
                dagster.AssetKey(temp_fname),
                {"source_path": temp_dir_name, "format": temp_extension},
            )
            self.assertIsInstance(result, pd.DataFrame)

    def test_throws_error_when_file_does_not_exist_and_no_url(self):
        with self.assertRaises(IOError):
            read_or_attempt_download(dagster.AssetKey("bad_file"), {"format": "csv"})

    def test_tries_to_download_when_url_specified(self):
        mock_download_func = create_autospec(download_uncached_file, return_value=True)
        try:
            asset_key = dagster.AssetKey("bad_file")
            metadata = {"format": "csv", "url": "badurl"}
            fpath = _get_fs_path(asset_key, metadata)
            read_or_attempt_download(asset_key, metadata)
            mock_download_func.assert_called_once_with("badurl", fpath)
        except IOError:
            pass



from typing import Dict, Tuple

import dagster
import pandas as pd

from ireiat.config.constants import (
    SUM_TONS_TOLERANCE,
    INTERMEDIATE_DIRECTORY_ARGS,
)
from ireiat.config.data_pipeline import FAF5MasterConfig, FAF5FilterConfig, FAF5DemandConfig
from ireiat.config.faf_enum import FAFMode
from ireiat.data_pipeline.assets.demand.faf5_helpers import (
    faf5_compute_county_tons_for_mode,
)
from ireiat.data_pipeline.metadata import publish_metadata


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def faf_filtered_grouped_tons(
    context: dagster.AssetExecutionContext,
    faf5_demand_src: pd.DataFrame,
    config: FAF5FilterConfig,
) -> pd.DataFrame:
    """Filters FAF by containerizable SCTG2 codes and relevant modes, multiplies by containerizable
    demand in each record, and groups by origin/destination/mode."""
    # limit to "containerizable" tons on relevant modes
    containerizable_codes = [x.sctg2 for x in config.faf_commodities if x.containerizable]
    context.log.info(f"Using {len(containerizable_codes)} containerizable codes")
    is_containerizable = faf5_demand_src["sctg2"].isin(containerizable_codes)
    is_relevant_mode = faf5_demand_src["dms_mode"].isin(
        [FAFMode.TRUCK, FAFMode.RAIL, FAFMode.WATER, FAFMode.MULTIPLE_AND_MAIL]
    )
    filtered_faf_pdf = faf5_demand_src.loc[is_containerizable & is_relevant_mode].copy()

    # reduce tons to what has been configured as "intermodal"
    intermodal_percentage_map = {
        x.sctg2: x.percentage_containerizable for x in config.faf_commodities
    }
    context.log.info(intermodal_percentage_map)
    for not_included_code in set(filtered_faf_pdf["sctg2"].unique()) - set(containerizable_codes):
        intermodal_percentage_map[not_included_code] = 1
        context.log.info(
            f"SCTG2 code {not_included_code} not included, assuming all is intermodal!"
        )
    intermodal_tons_percentages = filtered_faf_pdf["sctg2"].map(intermodal_percentage_map)
    filtered_faf_pdf[config.faf_demand_field] = (
        filtered_faf_pdf[config.faf_demand_field] * intermodal_tons_percentages
    )

    grouped_faf_pdf = filtered_faf_pdf.groupby(
        ["dms_orig", "dms_dest", "dms_mode"], as_index=False
    )[[config.faf_demand_field]].sum()
    non_zero_grouped_faf_pdf = grouped_faf_pdf.loc[grouped_faf_pdf[config.faf_demand_field] > 0]

    publish_metadata(context, non_zero_grouped_faf_pdf)
    return non_zero_grouped_faf_pdf


def _allocate_unknown_modes(
    context: dagster.AssetExecutionContext,
    entire_df: pd.DataFrame,
    specific_mode: int,
    percentage_unknown_to_specific: float,
    demand_field: str,
) -> pd.DataFrame:
    known_mode = entire_df.loc[entire_df["dms_mode"] == specific_mode]
    unknown_mode = entire_df.loc[entire_df["dms_mode"] == FAFMode.MULTIPLE_AND_MAIL.value]
    unknown_mode[demand_field] = unknown_mode[demand_field] * percentage_unknown_to_specific
    concat_pdf = pd.concat([known_mode, unknown_mode], axis=0)
    mode_pdf = concat_pdf.groupby(["dms_orig", "dms_dest"], as_index=False)[demand_field].sum()
    publish_metadata(context, mode_pdf)
    return mode_pdf


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def faf5_truck_demand(
    context: dagster.AssetExecutionContext,
    faf_filtered_grouped_tons: pd.DataFrame,
    config: FAF5DemandConfig,
) -> pd.DataFrame:
    """Aggregates FAF truck mode and relevant portion of unknown mode into a single dataframe,
    grouped by origin and destination, summing total tons"""
    return _allocate_unknown_modes(
        context,
        faf_filtered_grouped_tons,
        FAFMode.TRUCK.value,
        config.unknown_mode_percent,
        config.faf_demand_field,
    )


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def faf5_rail_demand(
    context: dagster.AssetExecutionContext,
    faf_filtered_grouped_tons: pd.DataFrame,
    config: FAF5DemandConfig,
) -> pd.DataFrame:
    """Aggregates FAF rail mode and relevant portion of unknown mode into a single dataframe,
    grouped by origin and destination, summing total tons"""
    return _allocate_unknown_modes(
        context,
        faf_filtered_grouped_tons,
        FAFMode.RAIL.value,
        config.unknown_mode_percent,
        config.faf_demand_field,
    )


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def faf5_water_demand(
    context: dagster.AssetExecutionContext,
    faf_filtered_grouped_tons: pd.DataFrame,
    config: FAF5DemandConfig,
) -> pd.DataFrame:
    """Aggregates FAF marine mode and relevant portion of unknown mode into a single dataframe,
    grouped by origin and destination, summing total tons"""
    return _allocate_unknown_modes(
        context,
        faf_filtered_grouped_tons,
        FAFMode.WATER.value,
        config.unknown_mode_percent,
        config.faf_demand_field,
    )


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def county_to_county_highway_tons(
    context: dagster.AssetExecutionContext,
    faf5_truck_demand: pd.DataFrame,
    faf_id_to_county_id_allocation_map: Dict[str, Dict[Tuple[str, str], float]],
    config: FAF5MasterConfig,
) -> pd.DataFrame:
    """Calculate (State FIPS origin, County FIPS origin), (State FIPS destination, County FIPS destination), tons
    for given mode based on county allocation percentages."""
    non_zero_county_od_pdf = faf5_compute_county_tons_for_mode(
        faf5_truck_demand,
        faf_id_to_county_id_allocation_map,
        config.faf_demand_field,
        SUM_TONS_TOLERANCE,
    )
    publish_metadata(context, non_zero_county_od_pdf)
    return non_zero_county_od_pdf


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def county_to_county_rail_tons(
    context: dagster.AssetExecutionContext,
    faf5_rail_demand: pd.DataFrame,
    faf_id_to_county_id_allocation_map: Dict[str, Dict[Tuple[str, str], float]],
    config: FAF5MasterConfig,
) -> pd.DataFrame:
    """Calculate (State FIPS origin, County FIPS origin), (State FIPS destination, County FIPS destination), tons
    for given mode based on county allocation percentages."""
    non_zero_county_od_pdf = faf5_compute_county_tons_for_mode(
        faf5_rail_demand,
        faf_id_to_county_id_allocation_map,
        config.faf_demand_field,
        SUM_TONS_TOLERANCE,
    )
    publish_metadata(context, non_zero_county_od_pdf)
    return non_zero_county_od_pdf


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def county_to_county_marine_tons(
    context: dagster.AssetExecutionContext,
    faf5_water_demand: pd.DataFrame,
    faf_id_to_county_id_allocation_map: Dict[str, Dict[Tuple[str, str], float]],
    config: FAF5MasterConfig,
) -> pd.DataFrame:
    """Calculate (State FIPS origin, County FIPS origin), (State FIPS destination, County FIPS destination), tons
    for given mode based on county allocation percentages."""
    non_zero_county_od_pdf = faf5_compute_county_tons_for_mode(
        faf5_water_demand,
        faf_id_to_county_id_allocation_map,
        config.faf_demand_field,
        SUM_TONS_TOLERANCE,
    )
    publish_metadata(context, non_zero_county_od_pdf)
    return non_zero_county_od_pdf



import dagster
import pandas as pd

from ireiat.util.graph import generate_ball_tree


@dagster.asset(io_manager_key="default_io_manager_intermediate_path")
def marine_ball_tree(
    marine_network_dataframe: pd.DataFrame,
):
    """BallTree for highway nodes from the marine network dataframe"""
    return generate_ball_tree(marine_network_dataframe)



from collections import defaultdict
from typing import Dict, Tuple

import dagster
import pandas as pd

from ireiat.config.constants import SUM_TONS_TOLERANCE


@dagster.asset(
    io_manager_key="default_io_manager_intermediate_path",
    description="(State FIPS, County FIPS) -> Metric",
)
def actual_state_county_to_metric_map(
    us_census_county_population_src: pd.DataFrame,
) -> Dict[Tuple[str, str], int]:
    result: Dict[Tuple[str, str], int] = dict()
    # ignore state total information
    county_only_data = us_census_county_population_src.loc[
        us_census_county_population_src["COUNTY"] != "000"
    ]
    for row in county_only_data.itertuples():
        result[(row.STATE, row.COUNTY)] = row.POPESTIMATE2022
    return result


@dagster.asset(
    io_manager_key="default_io_manager_intermediate_path",
    description="FAF -> (State FIPS, County FIPS) -> % allocation",
)
def faf_id_to_county_id_allocation_map(
    faf_id_to_county_areas: Dict[str, Dict[Tuple[str, str], float]],
    actual_state_county_to_metric_map: Dict[Tuple[str, str], int],
):

    faf_id_to_county_id_metric_map: Dict[str, Dict[Tuple[str, str], float]] = defaultdict(dict)

    # look up the county population within each faf zone
    for faf_id, vals in faf_id_to_county_areas.items():
        for (state_id, county_id), pct_area_county_in_faf in vals.items():
            total_county_pop = actual_state_county_to_metric_map[(state_id, county_id)]
            faf_id_to_county_id_metric_map[faf_id][(state_id, county_id)] = (
                total_county_pop * pct_area_county_in_faf
            )

    faf_total_metric_map = dict()
    for faf_id, vals_dict in faf_id_to_county_id_metric_map.items():
        faf_total_metric_map[faf_id] = sum(vals_dict.values())

    # check that the totals between the metric and those allocated to counties are equal
    assert sum(actual_state_county_to_metric_map.values()) == sum(faf_total_metric_map.values())

    # determine the percentage of the faf demand that should be allocated to the county
    # based on the allocation metric of interest
    faf_id_to_county_percent_map: Dict[str, Dict[Tuple[str, str], float]] = defaultdict(dict)
    for faf_id, vals in faf_id_to_county_id_metric_map.items():
        for (state_id, county_id), population_portion in vals.items():
            result_percent = population_portion / faf_total_metric_map[faf_id]
            faf_id_to_county_percent_map[faf_id][(state_id, county_id)] = result_percent

    # confirm that all the faf ids have a "total" allocation that sums to 1 (within some tolerance)
    assert all(
        [
            abs(sum(metric_map.values()) - 1) < SUM_TONS_TOLERANCE
            for metric_map in faf_id_to_county_percent_map.values()
        ]
    )

    return faf_id_to_county_percent_map



from collections import Counter
from itertools import chain
from typing import Dict, Tuple, List

import geopandas
import igraph as ig
import numpy as np
import pandas as pd
from sklearn.neighbors import BallTree

from ireiat.config.constants import LATLONG_CRS


def get_coordinates_from_geoframe(gdf: geopandas.GeoDataFrame) -> pd.DataFrame:
    """Given a GeoDataFrame of line segments, returns the origin and destination coordinates of each line segment
    in a pandas DataFrame. Returned fields are origin_latitude, origin_longitude, destination_latitude, destination_longitude.
    Will ensure that the CRS for lat/longs are equal to EPSG:4326"""
    if gdf.crs != LATLONG_CRS:
        gdf = gdf.to_crs(LATLONG_CRS)
    link_coords = (
        gdf.geometry.get_coordinates()
        .reset_index()
        .groupby("index")
        .agg({"x": ["first", "last"], "y": ["first", "last"]})
    )
    link_coords = link_coords.droplevel(0, axis=1)
    link_coords.columns = [
        "origin_longitude",
        "destination_longitude",
        "origin_latitude",
        "destination_latitude",
    ]
    link_coords = np.round(link_coords, 6)  # round to 6 decimal places of lat long
    assert len(link_coords) == len(gdf)
    return link_coords


def generate_zero_based_node_maps(link_coords: pd.DataFrame) -> Dict[Tuple[float, float], int]:
    """Returns a dictionary of (lat,long)->index based on each row in the dataframe.
    The final index in the dictionary represents the count-1 of the identified nodes.
    Note: Nodes are assumed to be able to be uniquely identified by lat/long but no checks are made on the
    precision of lat/long in the `link_coords` dataframe!"""

    latlong_node_idx_dict: Dict[Tuple[float, float], int] = dict()
    latlong_node_idx_counter = 0

    for row in link_coords.itertuples():
        origin_coords = (row.origin_latitude, row.origin_longitude)
        destination_coords = (row.destination_latitude, row.destination_longitude)
        if origin_coords not in latlong_node_idx_dict:
            latlong_node_idx_dict[origin_coords] = latlong_node_idx_counter
            latlong_node_idx_counter += 1
        if destination_coords not in latlong_node_idx_dict:
            latlong_node_idx_dict[destination_coords] = latlong_node_idx_counter
            latlong_node_idx_counter += 1
    return latlong_node_idx_dict


def get_allowed_node_indices(g: ig.Graph) -> List[int]:
    """Given a graph, checks its connectedness and returns the indices of the
    nodes of the largest connected component"""
    # filter for the largest connected component (we are basically throwing away the weakly connected nodes)
    connected_component_length = Counter([len(f) for f in g.connected_components()])
    biggest_connected_component = max(connected_component_length)

    print(f"Number of nodes in largest connected component {connected_component_length}")
    count_of_excluded_nodes = sum(
        [k * v for k, v in connected_component_length.items() if k != biggest_connected_component]
    )
    print(f"Excluded # of nodes that are not strongly connected {count_of_excluded_nodes}")
    allowed_node_indices = list(
        chain.from_iterable(
            [x for x in g.connected_components() if len(x) == biggest_connected_component]
        )
    )
    return allowed_node_indices


def generate_ball_tree(pdf: pd.DataFrame) -> BallTree:
    """Generates a ball tree assuming the data has the form tail|head|origin|destination"""
    # stack unique origins / destinations by lat/long
    origins = pdf[["tail", "origin_latitude", "origin_longitude"]].drop_duplicates()
    destinations = pdf[["head", "destination_latitude", "destination_longitude"]].drop_duplicates()
    origins = origins.rename(
        {"tail": "node", "origin_latitude": "latitude", "origin_longitude": "longitude"},
        axis="columns",
    )
    destinations = destinations.rename(
        {"head": "node", "destination_latitude": "latitude", "destination_longitude": "longitude"},
        axis="columns",
    )

    # order all nodes
    all_nodes = pd.concat([origins, destinations]).drop_duplicates().sort_values("node")
    # make sure the node order is continuous
    assert all_nodes["node"].max() == len(all_nodes) - 1
    node_lat_long_radians = np.deg2rad(np.array(all_nodes[["latitude", "longitude"]]))
    return BallTree(node_lat_long_radians, metric="haversine")



from typing import Dict, Tuple, Optional

import dagster
import pandas as pd

from ireiat.config.constants import EXCLUDED_FIPS_CODES_MAP, INTERMEDIATE_DIRECTORY_ARGS
from ireiat.config.data_pipeline import TAPFilterTonsConfig
from ireiat.data_pipeline.metadata import publish_metadata


def _generate_tons_dataframe(
    context: dagster.AssetExecutionContext,
    in_network_tons: pd.DataFrame,
    county_fips_to_network_node_idx: Dict[Tuple[str, str], int],
) -> pd.DataFrame:
    """Helper function to associate (STATE, COUNTY) tons to a target network node ID mapping"""
    od_tuples = []
    included_tons, excluded_tons = 0, 0
    for row in in_network_tons.itertuples():
        orig = (row.state_orig, row.county_orig)
        dest = (row.state_dest, row.county_dest)
        if (
            orig not in county_fips_to_network_node_idx
            or dest not in county_fips_to_network_node_idx
        ):
            excluded_tons += row.tons
            continue
        od_tuples.append(
            (
                county_fips_to_network_node_idx[orig],
                county_fips_to_network_node_idx[dest],
                row.tons,
            )
        )
        included_tons += row.tons

    total_tons = included_tons + excluded_tons
    context.log.info(
        f"Tons excluded {excluded_tons}, tons included {included_tons}: {excluded_tons / total_tons:.1%}"
    )

    trips = pd.DataFrame(od_tuples, columns=["from", "to", "tons"]).sort_values(["from", "to"])
    publish_metadata(context, trips)
    return trips


def _filter_tons_dataframe(
    context: dagster.AssetExecutionContext,
    tons_dataframe: pd.DataFrame,
    quantile_threshold: Optional[float] = None,
) -> pd.DataFrame:
    """Filters a state_orig | county_orig | state_dest | county_dest to exclude counties outside
    the continential US and to exclude counties with self-circulating flows. If `quantile_threshold`
    is specified, the dataframe is further filtered to include only tons above that quantile."""
    # eliminate states and territories we're not interested in
    relevant_county_ods = tons_dataframe.loc[
        ~(
            tons_dataframe["state_orig"].isin(EXCLUDED_FIPS_CODES_MAP.values())
            | tons_dataframe["state_dest"].isin(EXCLUDED_FIPS_CODES_MAP.values())
        )
    ]

    # eliminate "self-circulating" flows (same county -> same county)
    non_self_county_ods = relevant_county_ods.loc[
        ~(
            (relevant_county_ods["state_orig"] == relevant_county_ods["state_dest"])
            & (relevant_county_ods["county_orig"] == relevant_county_ods["county_dest"])
        )
    ]
    context.log.info(
        f"County ODs: {len(tons_dataframe):,}, Relevant ODs: {len(relevant_county_ods):,}, Non-self ODs {len(non_self_county_ods):,}"
    )
    context.log.info(
        f"County Tons: {tons_dataframe['tons'].sum():,.1f}, Relevant tons: {relevant_county_ods['tons'].sum():,.1f}, Non-self tons {non_self_county_ods['tons'].sum():,.1f}"
    )

    # start with some limit of ODs
    if quantile_threshold is None:
        return non_self_county_ods.reset_index(drop=True)

    tons_threshold = non_self_county_ods["tons"].quantile(quantile_threshold)
    subset_county_od = non_self_county_ods.loc[non_self_county_ods["tons"] > tons_threshold]

    subset_county_od = subset_county_od.sort_values(
        ["state_orig", "county_orig", "state_dest", "county_dest"]
    )
    context.log.info(subset_county_od["tons"].describe())
    context.log.info(subset_county_od["tons"].sum() / non_self_county_ods["tons"].sum())
    reindexed_subset_county_od = subset_county_od.reset_index(drop=True)
    publish_metadata(context, reindexed_subset_county_od)
    return reindexed_subset_county_od


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def tap_highway_tons(
    context: dagster.AssetExecutionContext,
    county_to_county_highway_tons: pd.DataFrame,
    county_fips_to_highway_network_node_idx: Dict[Tuple[str, str], int],
    config: TAPFilterTonsConfig,
) -> pd.DataFrame:
    """Tons attached to the highway network nodes (from, to, tons)"""
    in_network_highway_tons = _filter_tons_dataframe(
        context, county_to_county_highway_tons, quantile_threshold=config.quantile_threshold
    )
    return _generate_tons_dataframe(
        context, in_network_highway_tons, county_fips_to_highway_network_node_idx
    )


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def tap_marine_tons(
    context: dagster.AssetExecutionContext,
    county_to_county_marine_tons: pd.DataFrame,
    county_fips_to_marine_network_node_idx: Dict[Tuple[str, str], int],
    config: TAPFilterTonsConfig,
) -> pd.DataFrame:
    """Tons attached to the marine network nodes (from, to, tons)"""
    in_network_marine_tons = _filter_tons_dataframe(
        context, county_to_county_marine_tons, quantile_threshold=config.quantile_threshold
    )
    return _generate_tons_dataframe(
        context, in_network_marine_tons, county_fips_to_marine_network_node_idx
    )


@dagster.asset(
    io_manager_key="custom_io_manager",
    metadata={
        "format": "parquet",
        "write_kwargs": dagster.MetadataValue.json(
            {"index": False},
        ),
        **INTERMEDIATE_DIRECTORY_ARGS,
    },
)
def tap_rail_tons(
    context: dagster.AssetExecutionContext,
    county_to_county_rail_tons: pd.DataFrame,
    county_fips_to_rail_network_node_idx: Dict[Tuple[str, str], int],
    config: TAPFilterTonsConfig,
) -> pd.DataFrame:
    """Tons attached to the marine network nodes (from, to, tons)"""
    in_network_rail_tons = _filter_tons_dataframe(
        context, county_to_county_rail_tons, quantile_threshold=config.quantile_threshold
    )
    return _generate_tons_dataframe(
        context, in_network_rail_tons, county_fips_to_rail_network_node_idx
    )



import pandas as pd
from collections import defaultdict
from typing import Dict, Tuple


def faf5_compute_county_tons_for_mode(
    faf_demand_pdf: pd.DataFrame,
    faf_id_to_county_id_allocation_map: Dict[str, Dict[Tuple[str, str], float]],
    FAF_TONS_TARGET_FIELD: str,
    SUM_TONS_TOLERANCE: float,
) -> pd.DataFrame:
    """
    Compute county-to-county tons for a specific mode, distributing based on FAF zone to county allocation percentages.

    Args:
        faf_demand_pdf (pd.DataFrame): DataFrame containing demand data for a specific mode (truck, rail, water).
        faf_id_to_county_id_allocation_map (dict): Mapping from FAF zones to county allocation percentages.
        FAF_TONS_TARGET_FIELD (str): The column in the DataFrame that holds the tons of demand.
        SUM_TONS_TOLERANCE (float): Tolerance for checking the sum of tons.

    Returns:
        pd.DataFrame: DataFrame with non-zero tons, aggregated at the county-to-county level.
    """
    county_od: Dict[Tuple[str, str, str, str], float] = defaultdict(float)
    for row in faf_demand_pdf.itertuples():
        constituent_orig_counties_map = faf_id_to_county_id_allocation_map[row.dms_orig]
        constituent_dest_counties_map = faf_id_to_county_id_allocation_map[row.dms_dest]
        for (state_orig, county_orig), pct_in_county_orig in constituent_orig_counties_map.items():
            for (
                state_dest,
                county_dest,
            ), pct_in_county_dest in constituent_dest_counties_map.items():
                tons = getattr(row, FAF_TONS_TARGET_FIELD) * pct_in_county_orig * pct_in_county_dest
                county_od[(state_orig, county_orig, state_dest, county_dest)] += tons

    # Verify that the total tons sum within tolerance
    assert (
        abs(faf_demand_pdf[FAF_TONS_TARGET_FIELD].sum() - sum(county_od.values()))
        < SUM_TONS_TOLERANCE
    ), "Tons mismatch for mode."

    # Create a dataframe from the county_od dictionary
    county_od_pdf = pd.DataFrame(
        [(*k, v) for k, v in county_od.items()],
        columns=["state_orig", "county_orig", "state_dest", "county_dest", "tons"],
    )

    # Filter out rows with zero tons and sort
    non_zero_county_od_pdf = county_od_pdf.loc[county_od_pdf["tons"] > 0].sort_values("tons")

    return non_zero_county_od_pdf



from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Optional

from ireiat.config.constants import INTERMEDIATE_PATH, CACHE_PATH, RAW_PATH

intermediate_path = CACHE_PATH / INTERMEDIATE_PATH
raw_path = CACHE_PATH / RAW_PATH

default_highway_traffic_output_path = CACHE_PATH / "highway_traffic.parquet"
default_marine_traffic_output_path = CACHE_PATH / "marine_traffic.parquet"
default_rail_traffic_output_path = CACHE_PATH / "rail_traffic.parquet"


@dataclass
class RunConfig:
    """Simple run configuration allowing for default overrides in child classes or passed paths"""

    default_network_file_path: Path = None
    default_od_file_path: Path = None
    default_output_file_path: Path = None

    passed_network_file_path: Optional[Path] = None
    passed_od_file_path: Optional[Path] = None
    passed_output_file_path: Optional[Path] = None

    @property
    def network_file_path(self):
        return self.passed_network_file_path or self.default_network_file_path

    @property
    def od_file_path(self):
        return self.passed_od_file_path or self.default_od_file_path

    @property
    def output_file_path(self):
        return self.passed_output_file_path or self.default_output_file_path


marine_config = partial(
    RunConfig,
    default_network_file_path=intermediate_path / "tap_marine_network_dataframe.parquet",
    default_od_file_path=intermediate_path / "tap_marine_tons.parquet",
    default_output_file_path=default_marine_traffic_output_path,
)

highway_config = partial(
    RunConfig,
    default_network_file_path=intermediate_path / "tap_highway_network_dataframe.parquet",
    default_od_file_path=intermediate_path / "tap_highway_tons.parquet",
    default_output_file_path=default_highway_traffic_output_path,
)

rail_config = partial(
    RunConfig,
    default_network_file_path=intermediate_path / "tap_rail_network_dataframe.parquet",
    default_od_file_path=intermediate_path / "tap_rail_tons.parquet",
    default_output_file_path=default_rail_traffic_output_path,
)

run_config_map = {"marine": marine_config, "highway": highway_config, "rail": rail_config}


@dataclass
class PostprocessConfig:
    """Simple postprocess configuration allowing for default overrides in child classes or passed paths"""

    default_traffic_path: Path = None
    default_network_graph_path: Path = None
    default_shp_file_path: Path = None

    passed_traffic_path: Optional[Path] = None
    passed_network_graph_path: Optional[Path] = None
    passed_shp_file_path: Optional[Path] = None

    @property
    def traffic_file_path(self):
        return self.passed_traffic_path or self.default_traffic_path

    @property
    def network_graph_path(self):
        return self.passed_network_graph_path or self.default_network_graph_path

    @property
    def shp_file_path(self):
        return self.passed_shp_file_path or self.default_shp_file_path


highway_postprocess_config = partial(
    PostprocessConfig,
    default_traffic_path=default_highway_traffic_output_path,
    default_network_graph_path=intermediate_path / "strongly_connected_highway_graph",
    default_shp_file_path=raw_path / "faf5_highway_links.zip",
)

marine_postprocess_config = partial(
    PostprocessConfig,
    default_traffic_path=default_marine_traffic_output_path,
    default_network_graph_path=intermediate_path / "strongly_connected_marine_graph",
    default_shp_file_path=raw_path / "marine_links.zip",
)

rail_postprocess_config = partial(
    PostprocessConfig,
    default_traffic_path=default_rail_traffic_output_path,
    default_network_graph_path=intermediate_path / "rail_graph_with_county_connections",
    default_shp_file_path=raw_path / "narn_rail_links.zip",
)

postprocess_config_map = {
    "marine": marine_postprocess_config,
    "highway": highway_postprocess_config,
    "rail": rail_postprocess_config,
}



from enum import StrEnum


class VertexType(StrEnum):
    """
    Enum for vertex types in the rail network graph.
    """

    IM_TERMINAL = "Intermodal Terminal"
    IM_DUMMY_NODE = "Intermodal Dummy Node"
    COUNTY_CENTROID = "cc"


class EdgeType(StrEnum):
    """
    Enum for edge types in the rail network graph.
    """

    IM_CAPACITY = "Intermodal Capacity Link"
    IM_DUMMY = "Linkage between IM and rail"
    RAIL_LINK = "Rail Link"
    IMPEDANCE_LINK = "Impedance Link"
    COUNTY_TO_IM_LINK = "County to IM"
